from __future__ import annotations
from typing import Any, Callable
from bcps import core
from bcps.cli import dialog_creator, color, edits, save_management, main


class FeatureHandler:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file

    def get_features(self):
        cat_features = {"cats": edits.cat_editor.CatEditor.edit_cats}
        if core.core_data.config.get_bool(core.ConfigKey.SEPARATE_CAT_EDIT_OPTIONS):
            cat_features = {
                "unlock_remove_cats": edits.cat_editor.CatEditor.unlock_remove_cats_run,
                "upgrade_cats": edits.cat_editor.CatEditor.upgrade_cats_run,
                "true_form_remove_form_cats": edits.cat_editor.CatEditor.true_form_remove_form_cats_run,
                "force_true_form_cats": edits.cat_editor.CatEditor.force_true_form_cats_run,
                "fourth_form_remove_form_cats": edits.cat_editor.CatEditor.fourth_form_remove_form_cats_run,
                "force_fourth_form_cats": edits.cat_editor.CatEditor.force_fourth_form_cats_run,
                "upgrade_talents_remove_talents_cats": edits.cat_editor.CatEditor.upgrade_talents_remove_talents_cats_run,
                "unlock_remove_cat_guide": edits.cat_editor.CatEditor.unlock_cat_guide_remove_guide_run,
            }

        cat_features["special_skills"] = (
            edits.basic_items.BasicItems.edit_special_skills
        )

        features: dict[str, Any] = {
            "save_management": {
                "save_save": save_management.SaveManagement.save_save,
                "save_upload": save_management.SaveManagement.save_upload,
                "save_save_file": save_management.SaveManagement.save_save_dialog,
                "save_save_documents": save_management.SaveManagement.save_save_documents,
                "adb_push": save_management.SaveManagement.adb_push,
                "adb_push_rerun": save_management.SaveManagement.adb_push_rerun,
                "export_save": save_management.SaveManagement.export_save,
                "load_save": save_management.SaveManagement.load_save,
                "init_save": save_management.SaveManagement.init_save,
                "convert_region": save_management.SaveManagement.convert_save_cc,
                "convert_version": save_management.SaveManagement.convert_save_gv,
            },
            "items": {
                "catfood": edits.basic_items.BasicItems.edit_catfood,
                "xp": edits.basic_items.BasicItems.edit_xp,
                "normal_tickets": edits.basic_items.BasicItems.edit_normal_tickets,
                "rare_tickets": edits.basic_items.BasicItems.edit_rare_tickets,
                "rare_ticket_trade_feature_name": edits.rare_ticket_trade.RareTicketTrade.rare_ticket_trade,
                "platinum_tickets": edits.basic_items.BasicItems.edit_platinum_tickets,
                "legend_tickets": edits.basic_items.BasicItems.edit_legend_tickets,
                "platinum_shards": edits.basic_items.BasicItems.edit_platinum_shards,
                "np": edits.basic_items.BasicItems.edit_np,
                "leadership": edits.basic_items.BasicItems.edit_leadership,
                "battle_items": edits.basic_items.BasicItems.edit_battle_items,
                "catseyes": edits.basic_items.BasicItems.edit_catseyes,
                "catfruit": edits.basic_items.BasicItems.edit_catfruit,
                "talent_orbs": core.game.catbase.talent_orbs.SaveOrbs.edit_talent_orbs,
                "catamins": edits.basic_items.BasicItems.edit_catamins,
                "scheme_items": edits.basic_items.BasicItems.edit_scheme_items,
                "labyrinth_medals": edits.basic_items.BasicItems.edit_labyrinth_medals,
            },
            "cats_special_skills": cat_features,
            "levels": {
                "clear_tutorial": edits.clear_tutorial.clear_tutorial,
                "clear_story": core.game.map.story.StoryChapters.clear_story,
                "challenge_score": core.game.map.challenge.edit_challenge_score,
                "dojo_score": core.game.map.dojo.edit_dojo_score,
                "enigma": core.game.map.enigma.edit_enigma,
                "unlock_aku_realm": edits.aku_realm.unlock_aku_realm,
                "story_treasures": core.game.map.story.StoryChapters.edit_treasures,
                "outbreaks": core.game.map.outbreaks.Outbreaks.edit_outbreaks,
                "aku_chapters": core.game.map.aku.AkuChapters.edit_aku_chapters,
                "itf_timed_scores": core.game.map.story.StoryChapters.edit_itf_timed_scores,
                "filibuster_reclearing": edits.basic_items.BasicItems.allow_filibuster_stage_reclearing,
                "sol": core.game.map.event.EventChapters.edit_sol_chapters,
                "event": core.game.map.event.EventChapters.edit_event_chapters,
                "collab": core.game.map.event.EventChapters.edit_collab_chapters,
                "gauntlets": core.game.map.gauntlets.GauntletChapters.edit_gauntlets,
                "collab_gauntlets": core.game.map.gauntlets.GauntletChapters.edit_collab_gauntlets,
                "uncanny": core.game.map.uncanny.UncannyChapters.edit_uncanny,
                "behemoth_culling": core.game.map.gauntlets.GauntletChapters.edit_behemoth_culling,
                "legend_quest": core.game.map.legend_quest.LegendQuestChapters.edit_legend_quest,
                "towers": core.game.map.tower.TowerChapters.edit_towers,
                "zero_legends": core.game.map.zero_legends.ZeroLegendsChapters.edit_zero_legends,
            },
            "gamototo": {
                "engineers": edits.basic_items.BasicItems.edit_engineers,
                "base_materials": edits.basic_items.BasicItems.edit_base_materials,
                "gamatoto_xp_level": core.game.gamoto.gamatoto.edit_xp,
                "gamatoto_helpers": core.game.gamoto.gamatoto.edit_helpers,
                "ototo_cat_cannon": core.game.gamoto.ototo.edit_cannon,
                "cat_shrine": core.game.gamoto.cat_shrine.CatShrine.edit_catshrine,
            },
            "account": {
                "unban_account": save_management.SaveManagement.unban_account,
                "upload_items": save_management.SaveManagement.upload_items,
                "inquiry_code": edits.basic_items.BasicItems.edit_inquiry_code,
                "password_refresh_token": edits.basic_items.BasicItems.edit_password_refresh_token,
            },
            "gatya": {
                "rare_gatya_seed": edits.basic_items.BasicItems.edit_rare_gatya_seed,
                "normal_gatya_seed": edits.basic_items.BasicItems.edit_normal_gatya_seed,
                "event_gatya_seed": edits.basic_items.BasicItems.edit_event_gatya_seed,
            },
            "fixes": {
                "fix_gamatoto_crash": edits.fixes.Fixes.fix_gamatoto_crash,
                "fix_time_errors": edits.fixes.Fixes.fix_time_errors,
                "unlock_equip_menu": edits.basic_items.BasicItems.unlock_equip_menu,
                "fix_officer_pass_crash": core.OfficerPass.fix_crash,
            },
            "other": {
                "unlocked_slots": edits.basic_items.BasicItems.edit_unlocked_slots,
                "restart_pack": edits.basic_items.BasicItems.set_restart_pack,
                "special_skills": edits.basic_items.BasicItems.edit_special_skills,
                "playtime": core.game.catbase.playtime.edit,
                "enemy_guide": edits.enemy_editor.EnemyEditor.edit_enemy_guide,
                "user_rank_rewards": core.game.catbase.user_rank_rewards.edit_user_rank_rewards,
                "unlock_equip_menu": edits.basic_items.BasicItems.unlock_equip_menu,
                "gold_pass": core.game.catbase.nyanko_club.NyankoClub.edit_gold_pass,
                "medals": core.game.catbase.medals.Medals.edit_medals,
                "missions": core.game.catbase.mission.Missions.edit_missions,
            },
            "config": core.core_data.config.edit_config,
            "update_external": core.update_external_content,
            "exit": main.Main.exit_editor,
        }
        return features

    def get_feature(self, feature_name: str):
        feature_path = feature_name.split(".")
        feature_dict = self.get_features()
        feature = feature_dict
        for path in feature_path:
            feature = feature[path]

        return feature

    def search_features(
        self,
        name: str,
        parent_path: str = "",
        features: dict[str, Any] | None = None,
        found_features: set[str] | None = None,
    ) -> set[str]:
        name = name.lower().replace(" ", "")
        if features is None:
            features = self.get_features()
        if found_features is None:
            found_features = set()

        for feature_name_key, feature in features.items():
            feature_name = core.core_data.local_manager.get_key(feature_name_key)
            path = (
                f"{parent_path}.{feature_name_key}" if parent_path else feature_name_key
            )
            if isinstance(feature, dict):
                found_features.update(
                    self.search_features(
                        name,
                        path,
                        feature,  # type: ignore
                        found_features,
                    )
                )
            for alias in core.LocalManager.get_all_aliases(feature_name):
                if name in alias.lower().replace(" ", "") or name == "":
                    found_features.add(path)
                    break

        return found_features

    def display_features(self, features: list[str]):
        feature_names: list[str] = []
        for feature_name in features:
            feature_names.append(feature_name.split(".")[-1])
        print()
        dialog_creator.ListOutput(feature_names, [], "features", {}).display_locale(
            remove_alias=True
        )

    def select_features(self, features: list[str], parent_path: str = "") -> list[str]:
        if features != list(self.get_features().keys()):
            features.insert(0, "go_back")
        self.display_features(features)
        print()
        usr_input = color.ColoredInput().localize("select_features")
        selected_features: list[str] = []
        if usr_input.isdigit():
            usr_input = int(usr_input)
            if usr_input > len(features):
                color.ColoredText.localize("invalid_input")
            elif usr_input < 1:
                color.ColoredText.localize("invalid_input")
            else:
                feature_name_top = features[usr_input - 1]
                if feature_name_top == "go_back":
                    return list(self.get_features().keys())
                feature = self.get_feature(feature_name_top)
                if isinstance(feature, dict):
                    for feature_name in feature.keys():
                        feature_path = (
                            f"{parent_path}.{feature_name_top}.{feature_name}"
                            if parent_path
                            else f"{feature_name_top}.{feature_name}"
                        )
                        selected_features.append(feature_path)

                else:
                    feature_path = (
                        f"{parent_path}.{feature_name_top}"
                        if parent_path
                        else feature_name_top
                    )
                    selected_features.append(feature_path)

        else:
            selected_features = list(self.search_features(usr_input))

        return selected_features

    def select_features_run(self):
        features = self.get_features()
        features = list(features.keys())
        self.save_file.to_file_thread(self.save_file.get_temp_path())
        edits.clear_tutorial.clear_tutorial(self.save_file, False)
        self.save_file.show_ban_message = False

        while True:
            features = self.select_features(features)

            new_features: list[str] = []
            found_strs: list[str] = []
            for feature_ in features:
                if feature_.split(".")[-1] in found_strs:
                    continue
                found_strs.append(feature_.split(".")[-1])
                new_features.append(feature_)

            features = new_features
            feature = None
            if len(features) == 1:
                feature = features[0]
            if len(features) == 2 and features[0] == "go_back":
                feature = features[1]

            if not feature:
                continue

            feature = self.get_feature(feature)

            if isinstance(feature, Callable):
                self.do_save_actions()

                feature(self.save_file)

                self.save_file.to_file_thread(self.save_file.get_temp_path())

                features = self.get_features()
                features = list(features.keys())

                core.core_data.game_data_getter = None  # reset game data getter so that if an old version is removed, it will download the new version

    def do_save_actions(self):
        if core.core_data.config.get_bool(core.ConfigKey.CLEAR_TUTORIAL_ON_LOAD):
            edits.clear_tutorial.clear_tutorial(self.save_file, False)
        if core.core_data.config.get_bool(core.ConfigKey.REMOVE_BAN_MESSAGE_ON_LOAD):
            self.save_file.show_ban_message = False



from __future__ import annotations
from bcps import core


class GatyaItemNames:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.names = self.__get_names()

    def __get_names(self) -> list[str] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", "GatyaitemName.csv")
        if data is None:
            return None
        csv = core.CSV(data, core.Delimeter.from_country_code_res(self.save_file.cc))
        names: list[str] = []
        for line in csv:
            names.append(line[0].to_str())

        return names

    def get_name(self, index: int) -> str | None:
        if self.names is None:
            return None
        try:
            return self.names[index]
        except IndexError:
            return core.core_data.local_manager.get_key(
                "gatya_item_unknown_name", index=index
            )


class GatyaItemBuyItem:
    def __init__(
        self,
        id: int,
        rarity: int,
        reflect_or_storage: bool,
        price: int,
        stage_drop_id: int,
        quantity: int,
        server_id: int,
        category: int,
        index: int,
        src_item_id: int,
        main_menu_type: int,
        gatya_ticket_id: int,
        comment: str,
    ):
        self.id = id
        self.rarity = rarity
        self.reflect_or_storage = reflect_or_storage
        self.price = price
        self.stage_drop_id = stage_drop_id
        self.quantity = quantity
        self.server_id = server_id
        self.category = category
        self.index = index
        self.src_item_id = src_item_id
        self.main_menu_type = main_menu_type
        self.gatya_ticket_id = gatya_ticket_id
        self.comment = comment


class GatyaItemBuy:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.buy = self.get_buy()

    def get_buy(self) -> list[GatyaItemBuyItem] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("DataLocal", "Gatyaitembuy.csv")
        if data is None:
            return None
        csv = core.CSV(data)
        buy: list[GatyaItemBuyItem] = []
        for i, line in enumerate(csv.lines[1:]):
            try:
                buy.append(
                    GatyaItemBuyItem(
                        i,
                        line[0].to_int(),
                        line[1].to_bool(),
                        line[2].to_int(),
                        line[3].to_int(),
                        line[4].to_int(),
                        line[5].to_int(),
                        line[6].to_int(),
                        line[7].to_int(),
                        line[8].to_int(),
                        line[9].to_int(),
                        line[10].to_int(),
                        line[11].to_str(),
                    )
                )
            except IndexError:
                pass

        return buy

    def sort_by_index(self, items: list[GatyaItemBuyItem]):
        items.sort(key=lambda x: x.index)
        return items

    def get_by_category(self, category: int) -> list[GatyaItemBuyItem] | None:
        if self.buy is None:
            return None
        return self.sort_by_index(
            [item for item in self.buy if item.category == category]
        )



from __future__ import annotations

"""Module for handling logging"""
import traceback
from bcps import core
import time


class Logger:
    def __init__(self):
        """
        Initializes a Logger object
        """
        self.log_file = core.Path.get_documents_folder().add("bcps.log")
        self.log_data = self.log_file.read(True).split(b"\n")

    def get_time(self) -> str:
        """
        Returns the current time in the format: "HH:MM:SS"

        Returns:
            str: The current time
        """
        return time.strftime("%d/%m/%Y %H:%M:%S", time.localtime())

    def log_debug(self, message: str):
        """
        Logs a debug message

        Args:
            message (str): The message to log
        """
        self.log_data.append(core.Data(f"[DEBUG]::{self.get_time()} - {message}"))
        self.write()

    def log_info(self, message: str):
        """
        Logs an info message

        Args:
            message (str): The message to log
        """
        self.log_data.append(core.Data(f"[INFO]::{self.get_time()} - {message}"))
        self.write()

    def log_warning(self, message: str):
        """
        Logs a warning message

        Args:
            message (str): The message to log
        """
        self.log_data.append(core.Data(f"[WARNING]::{self.get_time()} - {message}"))
        self.write()

    def log_error(self, message: str):
        """
        Logs an error message

        Args:
            message (str): The message to log
        """
        self.log_data.append(core.Data(f"[ERROR]::{self.get_time()} - {message}"))
        self.write()

    def log_exception(self, exception: Exception, extra_msg: str = ""):
        tb = traceback.format_exc()
        if tb == "NoneType: None\n":
            try:
                raise exception
            except Exception:
                tb = traceback.format_exc()

        self.log_error(
            f"{extra_msg}: {exception.__class__.__name__}: {exception}\n{tb}"
        )

    def write(self):
        """
        Writes the log data to the log file
        """
        self.log_file.write(core.Data.from_many(self.log_data, core.Data("\n")).strip())

    def log_no_file_found(self, file_name: str):
        """
        Logs that a file was not found

        Args:
            fileName (str): The name of the file
        """
        self.log_warning(f"Could not find {file_name}")

    @staticmethod
    def get_traceback() -> str:
        """
        Gets the traceback of the last exception

        Returns:
            str: The traceback
        """
        tb = traceback.format_exc()
        if tb == "NoneType: None\n":
            return ""
        return tb



from __future__ import annotations
import random
from typing import Any
from bcps import core
from bcps.cli import color


class Upgrade:
    def __init__(self, plus: int, base: int):
        self.plus = plus
        self.base = base

        self.base_range = None
        self.plus_range = None

    def get_base(self) -> int:
        return self.base + 1

    def get_plus(self) -> int:
        return self.plus

    def upgrade(self):
        self.base += 1

    def increment_base(self, amount: int):
        self.base += amount

    def increment_plus(self, amount: int):
        self.plus += amount

    def get_random_base(self, max_base: int | None = None) -> int:
        if self.base_range is None:
            return self.base
        base = random.randint(self.base_range[0], self.base_range[1])
        if max_base is not None:
            base = min(base, max_base)
        return base

    def get_random_plus(self, max_plus: int | None = None) -> int:
        if self.plus_range is None:
            return self.plus
        plus = random.randint(self.plus_range[0], self.plus_range[1])
        if max_plus is not None:
            plus = min(plus, max_plus)
        return plus

    @staticmethod
    def read(stream: core.Data) -> Upgrade:
        plus = stream.read_ushort()
        base = stream.read_ushort()

        return Upgrade(plus, base)

    def write(self, stream: core.Data):
        stream.write_ushort(self.plus)
        stream.write_ushort(self.base)

    def serialize(self) -> dict[str, Any]:
        return {
            "plus": self.plus,
            "base": self.base,
        }

    @staticmethod
    def init() -> Upgrade:
        return Upgrade(0, 0)

    def reset(self):
        self.plus = 0
        self.base = 0

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Upgrade:
        return Upgrade(data.get("plus", 0), data.get("base", 0))

    def __repr__(self) -> str:
        return f"Upgrade(plus={self.plus}, base={self.base})"

    def __str__(self) -> str:
        return f"Upgrade(plus={self.plus}, base={self.base})"

    @staticmethod
    def get_user_upgrade(
        max_pos_base: int,
        max_pos_plus: int,
    ) -> tuple[Upgrade | None, bool]:
        color.ColoredText.localize(
            "max_upgrade", max_base=max_pos_base + 1, max_plus=max_pos_plus
        )
        usr_input = color.ColoredInput().localize("upgrade_input")
        if usr_input == core.core_data.local_manager.get_key("quit_key"):
            return None, True
        # example:
        # 10+20 = Upgrade(base=9, plus=20)
        # 10+ = Upgrade(base=9, plus=-1) # -1 means no change
        # +20 = Upgrade(base=-1, plus=20) # -1 means no change
        # 10 = Upgrade(base=9, plus=0)
        # 5-10+20-30 = Upgrade(base=random.randint(4, 9), plus=random.randint(20, 30))
        # 5-10+ = Upgrade(base=random.randint(4, 9), plus=-1)
        # +20-30 = Upgrade(base=-1, plus=random.randint(20, 30))
        # max+max = Upgrade(base=50000, plus=50000)

        parts = usr_input.split("+")
        if len(parts) == 1:
            base = parts[0]
            plus = "0"
        else:
            base = parts[0]
            plus = parts[1]

        min_base, max_base = None, None
        min_plus, max_plus = None, None

        max_text = core.core_data.local_manager.get_key("max")

        if not base:
            base_int = -1
        else:
            range_parts = base.split("-")
            if len(range_parts) == 1:
                if range_parts[0].strip() == max_text:
                    min_base = max_pos_base
                    max_base = max_pos_base
                else:
                    try:
                        min_base = int(range_parts[0]) - 1
                        max_base = min_base
                    except ValueError:
                        color.ColoredText.localize("invalid_upgrade_base", base=base)
                        return None, False
            else:
                try:
                    min_base = int(range_parts[0]) - 1
                    max_base = int(range_parts[1]) - 1
                except ValueError:
                    color.ColoredText.localize(
                        "invalid_upgrade_base_random",
                        min=range_parts[0],
                        max=range_parts[1],
                    )
                    return None, False

            base_int = (min_base + max_base) // 2

        if not plus:
            plus_int = -1
        else:
            range_parts = plus.split("-")
            if len(range_parts) == 1:
                if range_parts[0].strip() == max_text:
                    min_plus = max_pos_plus
                    max_plus = max_pos_plus
                else:
                    try:
                        min_plus = int(range_parts[0])
                        max_plus = min_plus
                    except ValueError:
                        color.ColoredText.localize("invalid_upgrade_plus", plus=plus)
                        return None, False
            else:
                try:
                    min_plus = int(range_parts[0])
                    max_plus = int(range_parts[1])
                except ValueError:
                    color.ColoredText.localize(
                        "invalid_upgrade_plus_random",
                        min=range_parts[0],
                        max=range_parts[1],
                    )
                    return None, False

            plus_int = (min_plus + max_plus) // 2

        upgrade = Upgrade(plus_int, base_int)
        upgrade.base_range = (min_base or base_int, max_base or base_int)
        upgrade.plus_range = (min_plus or plus_int, max_plus or plus_int)
        return upgrade, False

    def copy(self) -> Upgrade:
        upgrade = Upgrade(self.plus, self.base)
        upgrade.base_range = self.base_range
        upgrade.plus_range = self.plus_range
        return upgrade



from __future__ import annotations
from typing import Any, Callable
from bcps.cli import color

from bcps import core


class GameDataGetter:
    def __init__(self, cc: core.CountryCode):
        self.url = core.core_data.config.get_str(core.ConfigKey.GAME_DATA_REPO)
        self.lang = core.core_data.config.get_str(core.ConfigKey.LOCALE)
        self.cc = cc.get_cc_lang()
        self.real_cc = cc
        self.cc = self.cc if not self.cc.is_lang() else self.real_cc
        self.all_versions = self.get_versions()
        if self.all_versions is None:
            self.latest_version = None
        else:
            self.latest_version = self.get_latest_version(self.all_versions, self.cc)

    def does_save_version_match(self, save_file: core.SaveFile) -> bool:
        if self.latest_version is None:
            return False

        return save_file.game_version == self.latest_version[:-2]

    def get_latest_version(
        self, versions: list[str], cc: core.CountryCode
    ) -> str | None:
        length = len(versions)
        if cc == core.CountryCodeType.EN and length >= 1:
            return versions[0]
        if cc == core.CountryCodeType.JP and length >= 2:
            return versions[1]
        if cc == core.CountryCodeType.KR and length >= 3:
            return versions[2]
        if cc == core.CountryCodeType.TW and length >= 4:
            return versions[3]
        return None

    def get_versions(self) -> list[str] | None:
        response = core.RequestHandler(self.url + "latest.txt").get()
        if response is None:
            return None
        versions = response.text.split("\n")
        return versions

    def get_file(self, pack_name: str, file_name: str) -> core.Data | None:
        pack_name = self.get_packname(pack_name)
        version = self.latest_version
        if version is None:
            return None
        url = self.url + f"{version}/{pack_name}/{file_name}"
        response = core.RequestHandler(url).get()
        if response is None:
            return None
        if response.status_code != 200:
            return None
        return core.Data(response.content)

    def get_packname(self, packname: str) -> str:
        if packname != "resLocal":
            return packname
        if self.cc != core.CountryCodeType.EN:
            return packname
        langs = core.CountryCode.get_langs()
        if self.lang in langs:
            return f"{packname}_{self.lang}"
        return packname

    @staticmethod
    def get_game_data_dir() -> core.Path:
        return core.Path("game_data", is_relative=True)

    def get_file_path(self, pack_name: str, file_name: str) -> core.Path | None:
        version = self.latest_version

        if version is None:
            return None
        pack_name = self.get_packname(pack_name)
        path = GameDataGetter.get_game_data_dir().add(version).add(pack_name)
        path.generate_dirs()
        path = path.add(file_name)
        return path

    def save_file(self, pack_name: str, file_name: str) -> core.Data | None:
        pack_name = self.get_packname(pack_name)
        data = self.get_file(pack_name, file_name)
        if data is None:
            return None

        path = self.get_file_path(pack_name, file_name)
        if path is None:
            return None
        data.to_file(path)
        return data

    def is_downloaded(self, pack_name: str, file_name: str) -> bool:
        pack_name = self.get_packname(pack_name)
        path = self.get_file_path(pack_name, file_name)
        if path is None:
            return False
        return path.exists()

    def download_from_path(
        self, path: str, retries: int = 2, display_text: bool = True
    ) -> core.Data | None:
        pack_name, file_name = path.split("/")
        pack_name = self.get_packname(pack_name)
        return self.download(pack_name, file_name, retries, display_text)

    def download(
        self,
        pack_name: str,
        file_name: str,
        retries: int = 2,
        display_text: bool = True,
    ) -> core.Data | None:
        retries -= 1
        pack_name = self.get_packname(pack_name)

        if self.is_downloaded(pack_name, file_name):
            path = self.get_file_path(pack_name, file_name)
            if path is None:
                return None
            try:
                return path.read()
            except FileNotFoundError:
                return None

        if retries == 0:
            return None

        version = self.latest_version

        if version is None:
            if display_text:
                self.print_no_file(pack_name, file_name)
            return None

        if display_text:
            color.ColoredText.localize(
                "downloading",
                file_name=file_name,
                pack_name=pack_name,
                version=version,
            )
        data = self.save_file(pack_name, file_name)
        if data is None:
            data = self.download(pack_name, file_name, retries, display_text)
        if data is None:
            if display_text:
                self.print_no_file(pack_name, file_name)
            return None
        return data

    def download_all(
        self,
        pack_name: str,
        file_names: list[str],
        display_text: bool = True,
    ) -> list[tuple[str, core.Data] | None]:
        pack_name = self.get_packname(pack_name)

        callables: list[Callable[..., Any]] = []
        args: list[tuple[str, str, int, bool]] = []
        for file_name in file_names:
            callables.append(self.download)
            args.append((pack_name, file_name, 2, display_text))
        core.thread_run_many(callables, args)
        data_list: list[tuple[str, core.Data] | None] = []
        for file_name in file_names:
            path = self.get_file_path(pack_name, file_name)
            if path is None:
                data_list.append(None)
            elif not path.exists():
                data_list.append(None)
            else:
                data_list.append((file_name, path.read()))
        return data_list

    @staticmethod
    def get_all_downloaded_versions() -> list[str]:
        versions: list[str] = []
        for version in GameDataGetter.get_game_data_dir().get_dirs():
            if version.exists():
                versions.append(version.basename())
        return versions

    @staticmethod
    def delete_old_versions() -> None:
        versions = GameDataGetter.get_all_downloaded_versions()
        ccs: list[str] = []
        for version in versions:
            cc = version[-2:]
            if cc not in ccs:
                ccs.append(cc)

        for cc in ccs:
            versions_cc = list(filter(lambda x: x[-2:] == cc, versions))
            if len(versions_cc) <= 1:
                continue
            versions_cc.sort(reverse=True)
            for version in versions_cc[1:]:
                path = GameDataGetter.get_game_data_dir().add(version)
                path.remove()

    def print_no_file(self, packname: str, file_name: str) -> None:
        if self.latest_version is None:
            color.ColoredText.localize("failed_to_get_latest_version")
        else:
            color.ColoredText.localize(
                "failed_to_download_game_data",
                file_name=file_name,
                pack_name=packname,
                version=self.latest_version,
            )



from __future__ import annotations
import enum
from bcps.cli import dialog_creator
from bcps import core


class CountryCodeType(enum.Enum):
    EN = "en"
    JP = "jp"
    KR = "kr"
    TW = "tw"


class CountryCode:
    def __init__(self, cc: str | CountryCodeType):
        self.value = cc.value if isinstance(cc, CountryCodeType) else cc
        self.value = self.value.lower()

    def get_code(self) -> str:
        return self.value

    def get_client_info_code(self) -> str:
        code = self.get_code()
        if code == "jp":
            return "ja"
        return code

    def get_patching_code(self) -> str:
        code = self.get_code()
        if code == "jp":
            return ""
        return code

    @staticmethod
    def from_patching_code(code: str) -> CountryCode:
        if code == "":
            return CountryCode(CountryCodeType.JP)
        return CountryCode(code)

    @staticmethod
    def from_code(code: str) -> CountryCode:
        return CountryCode(code)

    @staticmethod
    def get_all() -> list["CountryCode"]:
        return [CountryCode(cc) for cc in CountryCodeType]

    @staticmethod
    def get_all_str() -> list[str]:
        ccts = CountryCode.get_all()
        return [cc.get_code() for cc in ccts]

    def __str__(self) -> str:
        return self.get_code()

    def __repr__(self) -> str:
        return self.get_code()

    def copy(self) -> CountryCode:
        return self

    @staticmethod
    def select() -> CountryCode | None:
        index = dialog_creator.ChoiceInput.from_reduced(
            CountryCode.get_all_str(),
            dialog="country_code_select",
            single_choice=True,
        ).single_choice()
        if index is None:
            return None
        return CountryCode.get_all()[index - 1]

    @staticmethod
    def select_from_ccs(ccs: list[CountryCode]) -> CountryCode | None:
        index = dialog_creator.ChoiceInput.from_reduced(
            [cc.get_code() for cc in ccs],
            dialog="country_code_select",
            single_choice=True,
        ).single_choice()
        if index is None:
            return None
        return ccs[index - 1]

    def __eq__(self, o: object) -> bool:
        if isinstance(o, CountryCode):
            return self.get_code() == o.get_code()
        elif isinstance(o, str):
            return self.get_code() == o
        elif isinstance(o, CountryCodeType):
            return self.get_code() == o.value
        return False

    def get_cc_lang(self) -> core.CountryCode:
        if core.core_data.config.get_bool(core.ConfigKey.FORCE_LANG_GAME_DATA):
            locale = core.core_data.config.get_str(core.ConfigKey.LOCALE)
            return core.CountryCode.from_code(locale)
        return self

    @staticmethod
    def get_langs() -> list[str]:
        return ["de", "it", "es", "fr", "th"]

    def is_lang(self) -> bool:
        return self.get_code() in CountryCode.get_langs()



from __future__ import annotations
from typing import Any
from bcps import core
from bcps.cli import color, dialog_creator
from bcps.cli.edits.cat_editor import SelectMode


class EnemyEditor:
    def __init__(self, save_file: core.SaveFile) -> None:
        self.save_file = save_file

    def unlock_enemy_guide(self, enemies: list[core.Enemy]):
        for enemy in enemies:
            enemy.unlock_enemy_guide(self.save_file)

        color.ColoredText.localize("unlock_enemy_guide_success")

    def remove_enemy_guide(self, enemies: list[core.Enemy]):
        for enemy in enemies:
            enemy.reset_enemy_guide(self.save_file)

        color.ColoredText.localize("remove_enemy_guide_success")

    def print_selected_enemies(self, enemies: list[core.Enemy]):
        if not enemies:
            return
        if len(enemies) > 50:
            color.ColoredText.localize("total_selected_enemies", total=len(enemies))
        else:
            for enemy in enemies:
                color.ColoredText.localize(
                    "selected_enemy",
                    id=enemy.id,
                    name=enemy.get_name(self.save_file),
                )

    def select(self, current_enemies: list[core.Enemy] | None):
        if current_enemies is None:
            current_enemies = []
        self.print_selected_enemies(current_enemies)

        options: dict[str, Any] = {
            "select_enemies_all": self.get_all_enemies,
            "select_enemies_id": self.select_id,
            "select_enemies_name": self.select_name,
        }
        option_id = dialog_creator.ChoiceInput.from_reduced(
            list(options), dialog="select_enemies", single_choice=True
        ).single_choice()
        if option_id is None:
            return current_enemies
        option_id -= 1

        func = options[list(options)[option_id]]
        new_enemies = func()
        if new_enemies is None:
            return None

        if current_enemies:
            mode_id = dialog_creator.IntInput().get_basic_input_locale("and_mode_q", {})
            if mode_id is None:
                mode = SelectMode.OR
            elif mode_id == 1:
                mode = SelectMode.AND
            elif mode_id == 2:
                mode = SelectMode.OR
            elif mode_id == 3:
                mode = SelectMode.REPLACE
            else:
                mode = SelectMode.OR
        else:
            mode = SelectMode.OR

        if mode == SelectMode.AND:
            return [enemy for enemy in new_enemies if enemy in current_enemies]
        if mode == SelectMode.OR:
            return list(set(current_enemies + new_enemies))
        if mode == SelectMode.REPLACE:
            return new_enemies
        return new_enemies

    def get_all_enemies(self) -> list[core.Enemy]:
        enemies: list[core.Enemy] = []
        for i in range(len(self.save_file.enemy_guide)):
            enemies.append(core.Enemy(i))
        return enemies

    def select_id(self) -> list[core.Enemy] | None:
        enemy_ids = dialog_creator.RangeInput(
            len(self.save_file.enemy_guide) - 1
        ).get_input_locale("enter_enemy_ids", {})
        if enemy_ids is None:
            return None
        enemy_ids = [enemy_id - 2 for enemy_id in enemy_ids]
        return self.get_enemies_by_id(enemy_ids)

    def get_enemies_by_id(self, ids: list[int]) -> list[core.Enemy]:
        enemies: list[core.Enemy] = []
        for enemy in self.get_all_enemies():
            if enemy.id in ids:
                enemies.append(enemy)
        return enemies

    def select_name(self) -> list[core.Enemy] | None:
        usr_name = dialog_creator.StringInput().get_input_locale("enter_enemy_name", {})
        if usr_name is None:
            return None
        enemies = self.get_enemies_by_name(usr_name)
        if not enemies:
            color.ColoredText.localize("enemy_not_found_name", name=usr_name)
            return None

        enemy_names = [enemy.get_name(self.save_file) for enemy in enemies]
        new_enemy_names: list[str] = []
        for enemy_name in enemy_names:
            if enemy_name is None:
                return None

            new_enemy_names.append(enemy_name)

        enemy_option_ids, _ = dialog_creator.ChoiceInput.from_reduced(
            new_enemy_names, dialog="select_enemies", single_choice=False
        ).multiple_choice()
        if enemy_option_ids is None:
            return None
        enemies_selected: list[core.Enemy] = []
        for enemy_option_id in enemy_option_ids:
            enemies_selected.append(enemies[enemy_option_id])
        return enemies_selected

    def get_enemies_by_name(self, name: str) -> list[core.Enemy]:
        enemies: list[core.Enemy] = []
        for enemy in self.get_all_enemies():
            enemy_name = enemy.get_name(self.save_file)
            if enemy_name is None:
                continue
            if name.lower() in enemy_name.lower():
                enemies.append(enemy)
        return enemies

    @staticmethod
    def from_save_file(
        save_file: core.SaveFile,
    ) -> tuple[EnemyEditor | None, list[core.Enemy]]:
        enemy_editor = EnemyEditor(save_file)
        current_enemies = enemy_editor.select([])
        if current_enemies is None:
            return None, []
        return enemy_editor, current_enemies

    @staticmethod
    def edit_enemy_guide(
        save_file: core.SaveFile,
        current_enemies: list[core.Enemy] | None = None,
        enemy_editor: EnemyEditor | None = None,
    ):
        if enemy_editor is None or current_enemies is None:
            enemy_editor, current_enemies = EnemyEditor.from_save_file(save_file)
        if enemy_editor is None or not current_enemies:
            return

        choice = dialog_creator.ChoiceInput.from_reduced(
            ["unlock_enemy_guide", "remove_enemy_guide"],
            dialog="edit_enemy_guide",
            single_choice=True,
        ).single_choice()
        if choice is None:
            return
        choice -= 1
        if choice == 0:
            enemy_editor.unlock_enemy_guide(current_enemies)
        elif choice == 1:
            enemy_editor.remove_enemy_guide(current_enemies)



from __future__ import annotations
from typing import Any
from bcps import core


class GameVersion:
    """A class to represent a game version."""

    def __init__(self, game_version: int):
        """Initializes a new instance of the GameVersion class.

        Args:
            game_version (int): Game version as an integer. e.g 120102 for 12.1.2
        """
        self.game_version = game_version

    def to_string(self) -> str:
        """Converts the game version to a string.

        Returns:
            str: Game version as a string. e.g 12.1.2
        """
        split_gv = str(self.game_version).zfill(6)
        split_gv = [str(int(split_gv[i : i + 2])) for i in range(0, len(split_gv), 2)]
        return ".".join(split_gv)

    def get_parts_zfill(self) -> list[str]:
        """Gets the parts of the game version as a list of strings with leading zeros.

        Returns:
            list[str]: Game version parts as strings with leading zeros. e.g ["12", "01", "02"]
        """
        return [part.zfill(2) for part in self.to_string().split(".")]

    def get_parts(self) -> list[int]:
        """Gets the parts of the game version as a list of integers.

        Returns:
            list[int]: Game version parts as integers. e.g [12, 1, 2]
        """
        return [int(part) for part in self.get_parts_zfill()]

    def format(self) -> str:
        """Formats the game version as a string with leading zeros.

        Returns:
            str: Game version as a string with leading zeros. e.g 12.01.02
        """
        parts = self.get_parts_zfill()
        string = ""
        for part in parts:
            string += f"{part}."
        return f"{string[:-1]}"

    def __str__(self) -> str:
        """Converts the game version to a string.

        Returns:
            str: Game version as a string. e.g 12.1.2
        """
        return self.to_string()

    def __repr__(self) -> str:
        """Converts the game version object to a string.

        Returns:
            str: Game version object as a string. e.g game_version(120102) 12.1.2
        """
        return f"game_version({self.game_version}) {self.to_string()}"

    @staticmethod
    def read(data: core.Data) -> GameVersion:
        """Reads a 4 byte int from a Data object.

        Args:
            data (core.Data): Data object to read from.

        Returns:
            GameVersion: Game version read from the Data object.
        """
        return GameVersion(data.read_int())

    def write(self, data: core.Data):
        """Writes the 4 byte game version to a Data object.

        Args:
            data (core.Data): Data object to write to.
        """
        data.write_int(self.game_version)

    def serialize(self) -> dict[str, Any]:
        """Serializes the game version to a dictionary.

        Returns:
            dict[str, Any]: Serialized game version.
        """
        return {"game_version": self.game_version}

    @staticmethod
    def deserialize(game_version: dict[str, Any]) -> GameVersion:
        """Deserializes a game version from a dictionary.

        Args:
            game_version (dict[str, Any]): Serialized game version.

        Returns:
            GameVersion: Deserialized game version.
        """
        return GameVersion(game_version["game_version"])

    @staticmethod
    def from_string(game_version: str) -> GameVersion:
        """Converts a string to a GameVersion object.

        Args:
            game_version (str): Game version as a string. e.g 12.1.2

        Returns:
            GameVersion: Game version as a GameVersion object.
        """
        split_gv = game_version.split(".")
        if len(split_gv) == 2:
            split_gv.append("0")
        final = ""
        for split in split_gv:
            final += split.zfill(2)
        return GameVersion(int(final))

    def __eq__(self, other: Any) -> bool:
        """Checks if the game version is equal to another object.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the game version is equal to the other object, False otherwise.
        """
        if isinstance(other, GameVersion):
            return self.game_version == other.game_version
        elif isinstance(other, int):
            return self.game_version == other
        elif isinstance(other, str):
            return self.game_version == GameVersion.from_string(other).game_version
        else:
            return False

    def __ne__(self, other: Any) -> bool:
        """Checks if the game version is not equal to another object.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the game version is not equal to the other object, False otherwise.
        """
        return not self.__eq__(other)

    def __lt__(self, other: Any) -> bool:
        """Checks if the game version is less than another object.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the game version is less than the other object, False otherwise.
        """
        if isinstance(other, GameVersion):
            return self.game_version < other.game_version
        elif isinstance(other, int):
            return self.game_version < other
        elif isinstance(other, str):
            return self.game_version < GameVersion.from_string(other).game_version
        else:
            return False

    def __le__(self, other: Any) -> bool:
        """Checks if the game version is less than or equal to another object.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the game version is less than or equal to the other object, False otherwise.
        """
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other: Any) -> bool:
        """Checks if the game version is greater than another object.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the game version is greater than the other object, False otherwise.
        """
        return not self.__le__(other)

    def __ge__(self, other: Any) -> bool:
        """Checks if the game version is greater than or equal to another object.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the game version is greater than or equal to the other object, False otherwise.
        """
        return not self.__lt__(other)

    def __add__(self, other: Any) -> GameVersion:
        """Adds the game version to another object.

        Args:
            other (Any): Object to add to.

        Returns:
            GameVersion: Game version added to the other object.
        """
        if isinstance(other, GameVersion):
            return GameVersion(self.game_version + other.game_version)
        elif isinstance(other, int):
            return GameVersion(self.game_version + other)
        elif isinstance(other, str):
            return GameVersion(
                self.game_version + GameVersion.from_string(other).game_version
            )
        else:
            return NotImplemented

    def __sub__(self, other: Any) -> GameVersion:
        """Subtracts the game version from another object.

        Args:
            other (Any): Object to subtract from.

        Returns:
            GameVersion: Game version subtracted from the other object.
        """
        return self.__add__(-other)



from __future__ import annotations
from typing import Any
from bcps import core


class MaxValueHelper:
    def __init__(self):
        self.max_value_data = self.get_max_value_data()

    def get_max_value_data(self) -> dict[str, Any]:
        file_path = core.Path("max_values.json", True)
        if not file_path.exists():
            return {}
        try:
            return core.JsonFile.from_data(file_path.read()).to_object()
        except core.JSONDecodeError:
            return {}

    def get(self, value_code: str) -> int:
        try:
            return int(self.max_value_data.get(value_code, 0))
        except ValueError:
            return 0

    def get_property(self, value_code: str, property: str) -> int:
        try:
            return int(self.max_value_data.get(value_code, {}).get(property, 0))
        except ValueError:
            return 0

    def get_old(self, value_code: str) -> int:
        return self.get_property(value_code, "old")

    def get_new(self, value_code: str) -> int:
        return self.get_property(value_code, "new")



from __future__ import annotations
from typing import Any
from bcps import core


class StampData:
    def __init__(
        self,
        current_stamp: int,
        collected_stamp: list[int],
        unknown: int,
        daily_reward: int,
    ):
        self.current_stamp = current_stamp
        self.collected_stamp = collected_stamp
        self.unknown = unknown
        self.daily_reward = daily_reward

    @staticmethod
    def init() -> StampData:
        return StampData(0, [0] * 30, 0, 0)

    @staticmethod
    def read(stream: core.Data) -> StampData:
        current_stamp = stream.read_int()
        collected_stamp = stream.read_int_list(30)
        unknown = stream.read_int()
        daily_reward = stream.read_int()
        return StampData(current_stamp, collected_stamp, unknown, daily_reward)

    def write(self, stream: core.Data):
        stream.write_int(self.current_stamp)
        stream.write_int_list(self.collected_stamp, write_length=False)
        stream.write_int(self.unknown)
        stream.write_int(self.daily_reward)

    def serialize(self) -> dict[str, Any]:
        return {
            "current_stamp": self.current_stamp,
            "collected_stamp": self.collected_stamp,
            "unknown": self.unknown,
            "daily_reward": self.daily_reward,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> StampData:
        return StampData(
            data.get("current_stamp", 0),
            data.get("collected_stamp", []),
            data.get("unknown", 0),
            data.get("daily_reward", 0),
        )

    def __repr__(self):
        return f"StampData({self.current_stamp}, {self.collected_stamp}, {self.unknown}, {self.daily_reward})"

    def __str__(self):
        return f"StampData({self.current_stamp}, {self.collected_stamp}, {self.unknown}, {self.daily_reward})"



from __future__ import annotations
from bcps import core

from bcps.cli import color, dialog_creator


class RareTicketTrade:
    @staticmethod
    def rare_ticket_trade(save_file: core.SaveFile):
        current_amount = save_file.rare_tickets
        max_amount = max(
            core.core_data.max_value_manager.get("rare_tickets") - current_amount, 0
        )
        if max_amount == 0:
            color.ColoredText.localize("rare_ticket_trade_maxed")
            return
        to_add = dialog_creator.IntInput(max_amount, 0).get_input_locale_while(
            "rare_ticket_trade_enter", {"max": max_amount, "current": current_amount}
        )
        if to_add is None:
            return

        space = False
        for storage_item in save_file.cats.storage_items:
            if storage_item.item_type == 0 or (
                storage_item.item_id == 1 and storage_item.item_type == 2
            ):
                storage_item.item_id = 1
                storage_item.item_type = 2
                space = True
                break

        if not space:
            color.ColoredText.localize("rare_ticket_trade_storage_full")
            return

        amount = to_add * 5
        save_file.gatya.trade_progress = amount

        color.ColoredText.localize(
            "rare_ticket_successfully_traded", rare_ticket_count=to_add
        )



from __future__ import annotations

from typing import Any
from bcps import core
from bcps.cli import dialog_creator


class BattleItem:
    def __init__(self, amount: int):
        self.amount = amount
        self.locked = False

    @staticmethod
    def init() -> BattleItem:
        return BattleItem(0)

    @staticmethod
    def read_amount(stream: core.Data) -> BattleItem:
        return BattleItem(stream.read_int())

    def write_amount(self, stream: core.Data):
        stream.write_int(self.amount)

    def read_locked(self, stream: core.Data):
        self.locked = stream.read_bool()

    def write_locked(self, stream: core.Data):
        stream.write_bool(self.locked)

    def serialize(self) -> dict[str, Any]:
        return {
            "amount": self.amount,
            "locked": self.locked,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> BattleItem:
        battle_item = BattleItem(data.get("amount", 0))
        battle_item.locked = data.get("locked", False)
        return battle_item

    def __repr__(self):
        try:
            return f"BattleItem({self.amount}, {self.locked})"
        except AttributeError:
            return f"BattleItem({self.amount})"

    def __str__(self):
        return self.__repr__()


class BattleItems:
    def __init__(self, items: list[BattleItem]):
        self.items = items
        self.lock_item = False

    @staticmethod
    def init() -> BattleItems:
        return BattleItems([BattleItem.init() for _ in range(6)])

    @staticmethod
    def read_items(stream: core.Data) -> BattleItems:
        total_items = 6
        items = [BattleItem.read_amount(stream) for _ in range(total_items)]
        return BattleItems(items)

    def write_items(self, stream: core.Data):
        for item in self.items:
            item.write_amount(stream)

    def read_locked_items(self, stream: core.Data):
        self.lock_item = stream.read_bool()
        for item in self.items:
            item.read_locked(stream)

    def write_locked_items(self, stream: core.Data):
        stream.write_bool(self.lock_item)
        for item in self.items:
            item.write_locked(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "items": [item.serialize() for item in self.items],
            "lock_item": self.lock_item,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> BattleItems:
        battle_items = BattleItems(
            [BattleItem.deserialize(item) for item in data.get("items", [])]
        )
        battle_items.lock_item = data.get("lock_item", False)
        return battle_items

    def __repr__(self):
        return f"BattleItems({self.items})"

    def __str__(self):
        return f"BattleItems({self.items})"

    def get_names(self, save_file: core.SaveFile) -> list[str] | None:
        names = core.core_data.get_gatya_item_names(save_file).names
        if names is None:
            return None
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(3)
        if items is None:
            return None

        names = [names[item.id] for item in items]
        return names

    def edit(self, save_file: core.SaveFile):
        group_name = save_file.get_localizable().get("shop_category1")
        if group_name is None:
            group_name = core.core_data.local_manager.get_key("battle_items")
        item_names = self.get_names(save_file)
        if item_names is None:
            return
        current_values = [item.amount for item in self.items]
        values = dialog_creator.MultiEditor.from_reduced(
            group_name,
            item_names,
            current_values,
            core.core_data.max_value_manager.get("battle_items"),
        ).edit()
        for i, value in enumerate(values):
            self.items[i].amount = value



from __future__ import annotations
import csv as csv_module
import enum
from typing import Any
import typing
from bcps import core


class DelimeterType(enum.Enum):
    COMMA = ","
    TAB = "\t"
    PIPE = "|"


class Delimeter:
    def __init__(self, de: DelimeterType | str):
        if isinstance(de, str):
            self.delimiter = DelimeterType(de)
        else:
            self.delimiter = de

    @staticmethod
    def from_country_code_res(cc: core.CountryCode) -> Delimeter:
        if cc.get_cc_lang() == core.CountryCodeType.JP:
            return Delimeter(DelimeterType.COMMA)
        return Delimeter(DelimeterType.PIPE)

    def __str__(self) -> str:
        return self.delimiter.value


class Cell:
    def __init__(self, dt: core.Data):
        self.data = dt

    def to_str(self) -> str:
        return self.data.to_str()

    def to_int(self) -> int:
        return self.data.to_int()

    def to_bool(self) -> bool:
        return self.data.to_bool()

    def __repr__(self) -> str:
        return f"Cell({self.data})"

    def __str__(self) -> str:
        return self.data.to_str()


class Row:
    def __init__(self, cells: list[Cell]):
        self.cells = cells
        self.index = 0

    @typing.overload
    def __getitem__(self, index: int) -> Cell: ...

    @typing.overload
    def __getitem__(self, index: slice) -> Row: ...

    def __getitem__(self, index: int | slice) -> Cell | Row:
        if isinstance(index, int):
            try:
                return self.cells[index]
            except IndexError:
                return Cell(core.Data(""))
        try:
            return Row(self.cells[index])
        except IndexError:
            return Row([])

    def __len__(self) -> int:
        return len(self.cells)

    @staticmethod
    def from_list(dt: list[core.Data]) -> Row:
        cells: list[Cell] = []
        for item in dt:
            cells.append(Cell(item))
        return Row(cells)

    def __repr__(self) -> str:
        return f"Row({self.cells})"

    def __str__(self) -> str:
        return self.__repr__()

    def __iter__(self):
        self.index = 0
        return iter(self.cells)

    def __next__(self):
        if self.index >= len(self.cells):
            raise StopIteration
        else:
            self.index += 1
            return self.cells[self.index - 1]

    def to_str_list(self) -> list[str]:
        return [cell.to_str() for cell in self.cells]

    def to_int_list(self) -> list[int]:
        return [cell.to_int() for cell in self.cells]


class CSV:
    def __init__(
        self,
        file_data: core.Data,
        delimiter: Delimeter | str = Delimeter(DelimeterType.COMMA),
        remove_padding: bool = True,
        remove_comments: bool = True,
        remove_empty: bool = True,
    ):
        self.file_data = file_data
        if remove_padding:
            data = self.file_data.unpad_pkcs7()
            if data is None:
                self.file_data = self.file_data
            else:
                self.file_data = data

        self.delimiter = delimiter
        self.remove_comments = remove_comments
        self.remove_empty = remove_empty
        self.index = 0
        self.parse()

    def parse(self):
        reader = csv_module.reader(
            self.file_data.data.decode("utf-8").splitlines(),
            delimiter=str(self.delimiter),
        )
        self.lines: list[Row] = []
        for row in reader:
            new_row: list[core.Data] = []
            full_row = f"{str(self.delimiter)}".join(row)
            if self.remove_comments:
                full_row = full_row.split("//")[0]
            row = full_row.split(str(self.delimiter))
            if row or not self.remove_empty:
                for item in row:
                    item = item.strip()
                    if item or not self.remove_empty:
                        new_row.append(core.Data(item))
                if new_row or not self.remove_empty:
                    self.lines.append(Row.from_list(new_row))

    def get_row(self, index: int) -> Row:
        try:
            return self.lines[index]
        except IndexError:
            return Row([])

    def __getitem__(self, index: int) -> Row:
        return self.get_row(index)

    def __len__(self) -> int:
        return len(self.lines)

    @staticmethod
    def from_file(
        pt: core.Path, delimiter: Delimeter = Delimeter(DelimeterType.COMMA)
    ) -> CSV:
        return CSV(pt.read(), delimiter)

    def add_line(self, line: list[Any] | Any):
        if not isinstance(line, list):
            line = [line]
        new_line: list[core.Data] = []
        for item in line:
            new_line.append(core.Data(str(item)))
        self.lines.append(Row.from_list(new_line))

    def set_line(self, index: int, line: list[Any]):
        new_line: list[core.Data] = []
        for item in line:
            new_line.append(core.Data(item))
        try:
            self.lines[index] = Row.from_list(new_line)
        except IndexError:
            self.lines.append(Row.from_list(new_line))

    def to_data(self) -> core.Data:
        csv: list[str] = []
        for line in self.lines:
            for i, item in enumerate(line):
                csv.append(str(item))
                if i != len(line) - 1:
                    csv.append(str(self.delimiter))
            csv.append("\r\n")
        return core.Data("".join(csv))

    def read_line(self) -> Row | None:
        try:
            line = self.lines[self.index]
        except IndexError:
            return None

        self.index += 1
        return line

    def reset_index(self):
        self.index = 0

    def has_line(self) -> bool:
        return self.index < len(self.lines)

    def __iter__(self):
        return self

    def __next__(self) -> Row:
        line = self.read_line()
        if line is None:
            raise StopIteration
        return line

    def extend(self, length: int, sub_length: int = 0):
        for _ in range(length):
            if sub_length == 0:
                self.lines.append(Row.from_list([]))
            else:
                self.lines.append(Row.from_list([core.Data("")] * sub_length))



from __future__ import annotations
from typing import Any, Callable

import bs4
from bcps import core
from bcps.cli import color


class MapNames:
    def __init__(self, save_file: core.SaveFile, code: str, output: bool = True):
        self.save_file = save_file
        self.out = output
        self.code = code
        self.gdg = core.core_data.get_game_data_getter(self.save_file)
        self.map_names: dict[int, str | None] = {}
        self.stage_names: dict[int, list[str]] = {}
        self.get_map_names()
        self.save_map_names()

    def get_file_path(self) -> core.Path:
        return (
            core.Path("map_names", True)
            .add(self.code)
            .generate_dirs()
            .add(f"{self.gdg.cc.get_code()}.json")
        )

    def read_map_names(self) -> dict[int, str | None]:
        file_path = self.get_file_path()
        if file_path.exists():
            names = core.JsonFile(file_path.read()).to_object()
            for id in names.keys():
                self.map_names[int(id)] = names[id]
            return names
        return {}

    def download_map_name(self, id: int):
        file_name = f"{self.code}{str(id).zfill(3)}.html"
        if self.gdg.cc != core.CountryCodeType.JP:
            url = f"https://ponosgames.com/information/appli/battlecats/stage/{self.gdg.cc.get_code()}/{file_name}"
        else:
            url = (
                f"https://ponosgames.com/information/appli/battlecats/stage/{file_name}"
            )
        data = core.RequestHandler(url).get()
        if data is None:
            return None
        if data.status_code != 200:
            return None
        html = data.content.decode("utf-8")
        bs = bs4.BeautifulSoup(html, "html.parser")
        name = bs.find("h2")
        if name is None:
            return None
        name = name.text.strip()
        if name:
            self.map_names[id] = name
        else:
            self.map_names[id] = None
        return name

    def get_map_names(self) -> dict[int, str | None] | None:
        names = self.read_map_names()
        gdg = core.core_data.get_game_data_getter(self.save_file)
        stage_names = gdg.download(
            "resLocal",
            f"StageName_R{self.code}_{core.core_data.get_lang(self.save_file)}.csv",
        )
        if stage_names is None:
            return None
        csv = core.CSV(
            stage_names,
            core.Delimeter.from_country_code_res(self.save_file.cc),
        )
        for i, row in enumerate(csv):
            stage_names_row = row.to_str_list()
            if not stage_names_row:
                continue
            self.stage_names[i] = stage_names_row

        total_downloaded = len(names)
        funcs: list[Callable[..., Any]] = []
        args: list[tuple[Any]] = []
        for i in range(len(csv)):
            if i < total_downloaded:
                continue
            funcs.append(self.download_map_name)
            args.append((i,))

        if self.out:
            color.ColoredText.localize("downloading_map_names", code=self.code)
        core.thread_run_many(funcs, args)
        return self.map_names

    def save_map_names(self):
        file_path = self.get_file_path()
        self.map_names = dict(sorted(self.map_names.items(), key=lambda item: item[0]))
        core.JsonFile.from_object(self.map_names).to_file(file_path)



from __future__ import annotations
from bcps import core


class Stage:
    def __init__(self, clear_amount: int):
        self.clear_amount = clear_amount

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read(stream: core.Data) -> Stage:
        clear_amount = stream.read_int()
        return Stage(clear_amount)

    def write(self, stream: core.Data):
        stream.write_int(self.clear_amount)

    def serialize(self) -> int:
        return self.clear_amount

    @staticmethod
    def deserialize(data: int) -> Stage:
        return Stage(data)

    def __repr__(self) -> str:
        return f"Stage(clear_amount={self.clear_amount!r})"

    def __str__(self) -> str:
        return f"Stage(clear_amount={self.clear_amount!r})"


class Chapter:
    def __init__(self, stages: list[Stage]):
        self.stages = stages

    @staticmethod
    def init() -> Chapter:
        return Chapter([Stage.init() for _ in range(12)])

    @staticmethod
    def read(stream: core.Data) -> Chapter:
        total = 12
        stages: list[Stage] = []
        for _ in range(total):
            stages.append(Stage.read(stream))
        return Chapter(stages)

    def write(self, stream: core.Data):
        for stage in self.stages:
            stage.write(stream)

    def serialize(self) -> list[int]:
        return [stage.serialize() for stage in self.stages]

    @staticmethod
    def deserialize(data: list[int]) -> Chapter:
        return Chapter([Stage.deserialize(stage) for stage in data])

    def __repr__(self) -> str:
        return f"Chapter(stages={self.stages!r})"

    def __str__(self) -> str:
        return f"Chapter(stages={self.stages!r})"


class ExChapters:
    def __init__(self, chapters: list[Chapter]):
        self.chapters = chapters

    @staticmethod
    def init() -> ExChapters:
        return ExChapters([])

    @staticmethod
    def read(stream: core.Data) -> ExChapters:
        total = stream.read_int()
        chapters: list[Chapter] = []
        for _ in range(total):
            chapters.append(Chapter.read(stream))

        return ExChapters(chapters)

    def write(self, stream: core.Data):
        stream.write_int(len(self.chapters))
        for chapter in self.chapters:
            chapter.write(stream)

    def serialize(self) -> list[list[int]]:
        return [chapter.serialize() for chapter in self.chapters]

    @staticmethod
    def deserialize(data: list[list[int]]) -> ExChapters:
        return ExChapters([Chapter.deserialize(chapter) for chapter in data])

    def __repr__(self) -> str:
        return f"Chapters(chapters={self.chapters!r})"

    def __str__(self) -> str:
        return f"Chapters(chapters={self.chapters!r})"



from __future__ import annotations
from typing import Any
from aenum import NamedConstant  # type: ignore
import colored  # type: ignore
from bcps import core


class ColorHex(NamedConstant):
    GREEN = "#008000"
    G = GREEN
    RED = "#FF0000"
    R = RED
    DARK_YELLOW = "#D7C32A"
    DY = DARK_YELLOW
    BLACK = "#000000"
    BL = BLACK
    WHITE = "#FFFFFF"
    W = WHITE
    CYAN = "#00FFFF"
    C = CYAN
    DARK_GREY = "#A9A9A9"
    DG = DARK_GREY
    BLUE = "#0000FF"
    B = BLUE
    YELLOW = "#FFFF00"
    Y = YELLOW
    MAGENTA = "#FF00FF"
    M = MAGENTA
    DARK_BLUE = "#00008B"
    DB = DARK_BLUE
    DARK_CYAN = "#008B8B"
    DC = DARK_CYAN
    DARK_MAGENTA = "#8B008B"
    DM = DARK_MAGENTA
    DARK_RED = "#8B0000"
    DR = DARK_RED
    DARK_GREEN = "#006400"
    DGN = DARK_GREEN
    LIGHT_GREY = "#D3D3D3"
    LG = LIGHT_GREY
    ORANGE = "#FFA500"
    O = ORANGE

    @staticmethod
    def from_name(name: str) -> str:
        if name == "":
            return ""
        return getattr(ColorHex, name.upper())


class ColorHelper:
    def __init__(self):
        self.theme_handler = core.core_data.theme_manager

    def get_color(self, color_name: str) -> str:
        try:
            first_char = color_name[0]
        except IndexError:
            return ""
        if first_char == "#":
            return color_name
        if first_char == "@":
            try:
                second_char = color_name[1]
            except IndexError:
                return ""
            try:
                third_char = color_name[2]
            except IndexError:
                third_char = ""
            if second_char == "p":
                return self.theme_handler.get_primary_color()
            if second_char == "s" and third_char != "u":
                return self.theme_handler.get_secondary_color()
            if second_char == "t":
                return self.theme_handler.get_tertiary_color()
            if second_char == "q":
                return self.theme_handler.get_quaternary_color()
            if second_char == "e":
                return self.theme_handler.get_error_color()
            if second_char == "w":
                return self.theme_handler.get_warning_color()
            if second_char == "s" and third_char == "u":
                return self.theme_handler.get_success_color()
            return self.theme_handler.get_theme_color(color_name[1:])
        try:
            return ColorHex.from_name(color_name)
        except AttributeError:
            return ""


class ColoredText:
    def __init__(self, string: str, end: str = "\n") -> None:
        string = string.replace("\\n", "\n")
        self.string = string
        self.end = end
        self.color_helper = ColorHelper()
        self.display(string)

    def display(self, string: str) -> None:
        text_data = self.parse(string)
        for i, (text, color) in enumerate(text_data):
            if i == len(text_data) - 1:
                text += self.end
            if color == "":
                print(text, end="")
            else:
                try:
                    fg = colored.fg(color)  # type: ignore
                except Exception:
                    print(text, end="")
                    continue
                print(colored.stylize(text, fg), end="")  # type: ignore

    @staticmethod
    def localize(string: str, escape: bool = True, **kwargs: Any) -> ColoredText:
        return ColoredText(
            core.core_data.local_manager.get_key(string, escape=escape, **kwargs)
        )

    def parse(self, txt: str) -> list[tuple[str, str]]:
        txt = "<@p>" + txt + "</>"
        output: list[tuple[str, str]] = []
        i = 0
        tags: list[str] = []
        inside_tag = False
        in_closing_tag = False
        tag_text = ""
        text = ""
        special_chars = core.LocalManager.get_special_chars()
        while i < len(txt):
            char = txt[i]
            if char == "\\" and i + 1 < len(txt) and txt[i + 1] in special_chars:
                i += 1
                char = txt[i]
                text += char
                i += 1
                continue
            if tags:
                tag = tags[-1]
            else:
                tag = ""
            if char == ">" and inside_tag:
                inside_tag = False
                if not in_closing_tag:
                    tags.append(tag_text)
                if in_closing_tag:
                    in_closing_tag = False
                tag_text = ""
            if char == "<" and not inside_tag:
                inside_tag = True
                if text:
                    color = self.color_helper.get_color(tag)
                    output.append((text, color))
                    text = ""
                    tag_text = ""
            if char == "/" and inside_tag:
                in_closing_tag = True
                if tags:
                    tags.pop()
            if not inside_tag and char != ">" and char != "<":
                text += char
            if inside_tag and char != "<" and char != ">":
                tag_text += char
            i += 1
        return output


class ColoredInput:
    def __init__(self, end: str = "") -> None:
        self.end = end

    def get(self, display_string: str) -> str:
        ColoredText(display_string, end=self.end)
        return input()

    def localize(self, string: str, escape: bool = True, **kwargs: Any) -> str:
        text = core.core_data.local_manager.get_key(string, escape=escape, **kwargs)
        return self.get(text)



from __future__ import annotations
from dataclasses import dataclass

from bcps import core
from bcps.cli import color, dialog_creator


@dataclass
class PlayTime:
    frames: int

    @staticmethod
    def get_fps() -> int:
        return 30

    @property
    def seconds(self) -> int:
        return self.frames // self.get_fps()

    @property
    def minutes(self) -> int:
        return self.seconds // 60

    @property
    def hours(self) -> int:
        return self.minutes // 60

    @property
    def just_seconds(self) -> int:
        return self.seconds % 60

    @property
    def just_minutes(self) -> int:
        return self.minutes % 60

    @property
    def just_hours(self) -> int:
        return self.hours % 60

    @staticmethod
    def from_hours(hours: int) -> PlayTime:
        return PlayTime(hours * 60 * 60 * PlayTime.get_fps())

    @staticmethod
    def from_minutes(minutes: int) -> PlayTime:
        return PlayTime(minutes * 60 * PlayTime.get_fps())

    @staticmethod
    def from_seconds(seconds: int) -> PlayTime:
        return PlayTime(seconds * PlayTime.get_fps())

    @staticmethod
    def from_hours_mins_secs(hours: int, minutes: int, seconds: int) -> PlayTime:
        return (
            PlayTime.from_hours(hours)
            + PlayTime.from_minutes(minutes)
            + PlayTime.from_seconds(seconds)
        )

    def __add__(self, other: PlayTime) -> PlayTime:
        return PlayTime(self.frames + other.frames)


def edit(save_file: core.SaveFile):
    play_time = PlayTime(save_file.officer_pass.play_time)
    color.ColoredText.localize(
        "playtime_current",
        hours=play_time.hours,
        minutes=play_time.just_minutes,
        seconds=play_time.just_seconds,
        frames=play_time.frames,
    )
    hours, _ = dialog_creator.IntInput().get_input("playtime_hours_prompt", {})
    if hours is None:
        return
    minutes, _ = dialog_creator.IntInput().get_input("playtime_minutes_prompt", {})
    if minutes is None:
        return
    seconds, _ = dialog_creator.IntInput().get_input("playtime_seconds_prompt", {})
    if seconds is None:
        return

    play_time = PlayTime.from_hours_mins_secs(hours, minutes, seconds)
    save_file.officer_pass.play_time = play_time.frames
    color.ColoredText.localize(
        "playtime_edited",
        hours=play_time.hours,
        minutes=play_time.just_minutes,
        seconds=play_time.just_seconds,
        frames=play_time.frames,
    )



from __future__ import annotations
import dataclasses
import tempfile
from typing import Any
from bcps import core
from bcps.cli import color


class PropertySet:
    """Represents a set of properties in a property file."""

    def __init__(self, locale: str, property: str):
        """Initializes a new instance of the PropertySet class.

        Args:
            locale (str): Language code of the locale.
            property (str): Name of the property file.
        """
        self.locale = locale
        self.path = LocalManager.get_locale_folder(locale).add(property + ".properties")
        self.properties: dict[str, str] = {}
        self.parse()

    def parse(self):
        """Parses the property file.

        Raises:
            KeyError: If a key is already defined in the property file.
        """
        lines = self.path.read().to_str().splitlines()
        i = 0
        in_multi_line = False
        multi_line_text = ""
        multi_line_key = ""
        while i < len(lines):
            line = lines[i]
            if line.startswith("#") or not line:
                i += 1
                continue
            if line.startswith(">") and in_multi_line:
                multi_line_text += line[1:] + "\n"
            if (in_multi_line and not line.startswith(">")) or (
                in_multi_line and i == len(lines) - 1
            ):
                in_multi_line = False
                if multi_line_key in self.properties:
                    raise KeyError(
                        f"Key {multi_line_key} already exists in property file"
                    )
                self.properties[multi_line_key] = multi_line_text[:-1]
                multi_line_text = ""
                multi_line_key = ""

            parts = line.split("=")
            if line.strip().endswith("="):
                in_multi_line = True
                multi_line_key = parts[0]

            if not in_multi_line:
                key = parts[0]
                value = "=".join(parts[1:])
                if key in self.properties:
                    raise KeyError(f"Key {key} already exists in property file")
                self.properties[key] = value

            i += 1

    def get_key(self, key: str) -> str:
        """Gets a key from the property file.

        Args:
            key (str): Key to get.

        Returns:
            str: Value of the key.
        """
        return self.properties.get(key, key).replace("\\n", "\n").replace("\\t", "\t")

    @staticmethod
    def from_config(property: str) -> PropertySet:
        """Gets a PropertySet from the language code in the config.

        Args:
            property (str): Name of the property file.

        Returns:
            PropertySet: PropertySet for the property file.
        """
        return PropertySet(
            core.core_data.config.get_str(core.ConfigKey.LOCALE), property
        )


class LocalManager:
    """Manages properties for a locale"""

    def __init__(self, locale: str | None = None):
        """Initializes a new instance of the LocalManager class.

        Args:
            locale (str): Language code of the locale.
        """
        if locale is None:
            lc = core.core_data.config.get_str(core.ConfigKey.LOCALE)
        else:
            lc = locale

        self.locale = lc
        self.path = LocalManager.get_locale_folder(lc)
        self.properties: dict[str, PropertySet] = {}
        self.all_properties: dict[str, str] = {}
        self.en_properties: dict[str, str] = {}
        self.en_properties_path = LocalManager.get_locale_folder("en")
        self.parse()
        if self.locale == "en":
            self.en_properties = self.all_properties

    def parse(self):
        """Parses all property files in the locale folder recursively."""
        for file in self.path.glob("**/*.properties", recursive=True):
            file_name = file.strip_path_from(self.path).path
            property_set = PropertySet(self.locale, file_name[:-11])
            self.all_properties.update(property_set.properties)
            self.properties[file_name[:-11]] = property_set

        if self.locale != "en":
            for file in self.en_properties_path.glob("**/*.properties", recursive=True):
                file_name = file.strip_path_from(self.en_properties_path).path
                property_set = PropertySet("en", file_name[:-11])
                self.en_properties.update(property_set.properties)

    def get_key(self, key: str, escape: bool = True, **kwargs: Any) -> str:
        """Gets a key from the property file.

        Args:
            key (str): Key to get.

        Returns:
            str: Value of the key.
        """
        try:
            text = self.get_key_recursive(key, kwargs, escape)
        except RecursionError:
            text = key

        for kwarg_key, kwarg_value in kwargs.items():
            value = str(kwarg_value)
            if escape:
                value = LocalManager.escape_string(value)
            text = text.replace("{" + kwarg_key + "}", value)

        if "$(" in text:
            text = self.parse_condition(text, kwargs)

        return text

    def parse_condition(self, text: str, kwargs: dict[str, Any]) -> str:
        counter = 0
        final_text = ""
        in_expression = False
        expression_text = ""
        count_down = 0
        while counter < len(text):
            char = text[counter]
            if counter == len(text) - 1:
                final_text += char
                break
            next_char = text[counter + 1]
            if char == "\\":
                final_text += next_char
                counter += 2
                continue
            if char == "$" and next_char == "(":
                count_down = 0
                in_expression = True
            elif char == "/" and next_char == "$":
                count_down = 2
                in_expression = False
                if len(expression_text) < 3:
                    counter += 1
                    continue
                new_expression_text = expression_text[2:-1]
                expression_text = ""
                parts = new_expression_text.split(":")
                if len(parts) < 2:
                    counter += 1
                    continue
                keyword = parts[0].strip()
                expression = parts[1].strip()
                conditions = expression.split("$,")
                string = ""
                for i, condition in enumerate(conditions):
                    condition = condition.strip()
                    if not condition:
                        continue
                    if i == len(conditions) - 1:
                        string = condition
                        break
                    condition_parts = condition.split("($")
                    if len(condition_parts) < 2:
                        continue
                    logic = condition_parts[0].strip()
                    word = condition_parts[1].strip()
                    if not word:
                        continue
                    word = word[:-1]
                    value = kwargs.get(keyword)
                    if value is None:
                        continue
                    equality = None
                    if logic.startswith("=="):
                        equality = "=="
                    elif logic.startswith("!="):
                        equality = "!="
                    elif logic.startswith(">="):
                        equality = ">="
                    elif logic.startswith("<="):
                        equality = "<="
                    elif logic.startswith(">"):
                        equality = ">"
                    elif logic.startswith("<"):
                        equality = "<"
                    if equality is None:
                        continue
                    logic_parts = logic.split(equality)
                    if len(logic_parts) < 2:
                        continue
                    logic_value = logic_parts[1].strip()

                    if isinstance(value, int):
                        if not logic_value.isdigit():
                            continue
                        logic_value = int(logic_value)

                    if equality == "==":
                        if logic_value == value:
                            string = word
                            break
                    elif equality == "!=":
                        if logic_value != value:
                            string = word
                            break

                    if isinstance(logic_value, int) and not string:
                        if equality == ">":
                            if logic_value > value:
                                string = word
                                break
                        elif equality == ">=":
                            if logic_value >= value:
                                string = word
                                break
                        elif equality == "<":
                            if logic_value < value:
                                string = word
                                break
                        elif equality == "<=":
                            if logic_value <= value:
                                string = word
                                break

                final_text += string

            if in_expression:
                expression_text += char
            else:
                if count_down <= 0:
                    final_text += char
                else:
                    count_down -= 1

            counter += 1

        return final_text

    @staticmethod
    def get_special_chars() -> list[str]:
        return ["<", ">", "/"]

    @staticmethod
    def escape_string(string: str) -> str:
        for char in LocalManager.get_special_chars():
            string = string.replace(char, "\\" + char)
        return string

    def get_key_recursive(
        self,
        key: str,
        kwargs: dict[str, Any],
        escape: bool = True,
    ) -> str:
        value = self.all_properties.get(key)
        if value is None:
            value = self.en_properties.get(key, key)
        value = value.replace("\\n", "\n").replace("\\t", "\t")
        # replace {{key}} with the value of the key
        if "{{" not in value:
            return value
        char_index = 0
        while char_index < len(value):
            if value[char_index] == "{" and value[char_index + 1] == "{":
                key_name = ""
                char_index += 2
                while value[char_index] != "}":
                    key_name += value[char_index]
                    char_index += 1

                if key_name != key:
                    value = value.replace(
                        "{{" + key_name + "}}", self.get_key(key_name, escape, **kwargs)
                    )
            char_index += 1

        return value

    @staticmethod
    def get_all_aliases(value: str) -> list[str]:
        """Gets all aliases from a string. Aliases are separated by |.

        Args:
            value (str): String to get aliases from.

        Returns:
            list[str]: List of aliases.
        """
        if "|" not in value:
            return [value]
        i = 0
        aliases: list[str] = []
        while i < len(value):
            char = value[i]
            prev_char = value[i - 1] if i > 0 else ""
            if char == "|" and prev_char != "\\":
                aliases.append(value[:i])
                value = value[i + 1 :]
                i = 0
            i += 1

        aliases.append(value)
        return aliases

    @staticmethod
    def from_config() -> LocalManager:
        """Gets a LocalManager from the language code in the config.

        Returns:
            LocalManager: LocalManager for the locale.
        """
        return LocalManager(core.core_data.config.get_str(core.ConfigKey.LOCALE))

    def check_duplicates(self):
        """Checks for duplicate keys in all property files.

        Raises:
            KeyError: If a key is already defined in the property file.
        """
        keys: set[str] = set()
        for property in self.properties.values():
            for key in property.properties.keys():
                if key in keys:
                    raise KeyError(f"Duplicate key {key}")
                keys.add(key)

    @staticmethod
    def get_all_locales() -> list[str]:
        """Gets all locales in the locales folder.

        Returns:
            list[str]: List of locales.
        """
        locales: list[str] = []
        for folder in LocalManager.get_locales_folder().get_dirs():
            locales.append(folder.basename())
        for folder in LocalManager.get_external_locales_folder().get_dirs():
            locales.append(folder.basename())
        return locales

    @staticmethod
    def get_locales_folder() -> core.Path:
        """Gets the locales folder.

        Returns:
            core.Path: Path to the locales folder.
        """
        return core.Path("locales", True)

    @staticmethod
    def get_external_locales_folder() -> core.Path:
        """Gets the external locales folder.

        Returns:
            core.Path: Path to the external locales folder.
        """
        return core.Path.get_documents_folder().add("external_locales")

    @staticmethod
    def get_locale_folder(locale: str) -> core.Path:
        """Gets the folder for a locale.

        Args:
            locale (str): Language code of the locale.

        Returns:
            core.Path: Path to the locale folder.
        """
        if locale.startswith("ext-"):
            return LocalManager.get_external_locales_folder().add(locale)
        return LocalManager.get_locales_folder().add(locale)

    @staticmethod
    def remove_locale(locale: str):
        """Removes a locale.

        Args:
            locale (str): Language code of the locale.
        """
        if locale not in LocalManager.get_all_locales():
            return
        if locale.startswith("ext-"):
            extern = ExternalLocaleManager.get_external_locale(locale)
            if extern is not None:
                ExternalLocaleManager.delete_locale(extern)
            LocalManager.get_external_locales_folder().add(locale).remove()
        else:
            LocalManager.get_locales_folder().add(locale).remove()

        if core.core_data.config.get_str(core.ConfigKey.LOCALE) == locale:
            core.core_data.config.set(core.ConfigKey.LOCALE, "en")


@dataclasses.dataclass
class ExternalLocale:
    short_name: str
    name: str
    description: str
    author: str
    version: str
    git_repo: str | None = None

    def to_json(self) -> dict[str, Any]:
        return dataclasses.asdict(self)

    @staticmethod
    def from_json(json_data: dict[str, Any]) -> ExternalLocale | None:
        short_name = json_data.get("short_name")
        name = json_data.get("name")
        description = json_data.get("description")
        author = json_data.get("author")
        version = json_data.get("version")
        git_repo = json_data.get("git_repo")
        if (
            short_name is None
            or name is None
            or description is None
            or author is None
            or version is None
        ):
            return None
        return ExternalLocale(
            short_name,
            name,
            description,
            author,
            version,
            git_repo,
        )

    @staticmethod
    def from_git_repo(git_repo: str) -> ExternalLocale | None:
        repo = core.GitHandler().get_repo(git_repo)
        if repo is None:
            return None
        locale_json = repo.get_file(core.Path("locale.json"))
        if locale_json is None:
            return None
        json_data = core.JsonFile.from_data(locale_json).to_object()
        json_data["git_repo"] = git_repo
        return ExternalLocale.from_json(json_data)

    def get_new_version(self) -> bool:
        if self.git_repo is None:
            return False
        repo = core.GitHandler().get_repo(self.git_repo)
        if repo is None:
            return False
        with tempfile.TemporaryDirectory() as tmp:
            temp_dir = core.Path(tmp)
            success = repo.clone_to_temp(temp_dir)
            if not success:
                return False
            external_locale = ExternalLocaleManager.parse_external_locale(temp_dir)
            if external_locale is None:
                return False
            version = external_locale.version

            if version == self.version:
                return False

            self.name = external_locale.name
            self.short_name = external_locale.short_name
            self.description = external_locale.description
            self.author = external_locale.author
            self.version = version

        success = repo.pull()
        if not success:
            return False
        self.save()
        return True

    def save(self):
        ExternalLocaleManager.save_locale(self)

    def get_full_name(self) -> str:
        return f"ext-{self.author}-{self.short_name}"


class ExternalLocaleManager:
    @staticmethod
    def delete_locale(external_locale: ExternalLocale):
        if external_locale.git_repo is None:
            return
        folder = core.GitHandler.get_repo_folder().add(
            external_locale.git_repo.split("/")[-1]
        )
        folder.remove()

    @staticmethod
    def save_locale(
        external_locale: ExternalLocale,
    ):
        """Saves an external locale.

        Args:
            external_locale (ExternalLocale): External locale to save.
        """
        if external_locale.git_repo is None:
            return
        folder = LocalManager.get_external_locales_folder().add(
            external_locale.get_full_name()
        )
        folder.generate_dirs()

        repo = core.GitHandler().get_repo(external_locale.git_repo)
        if repo is None:
            return
        files_dir = repo.get_folder(core.Path("files"))
        if files_dir is None:
            return

        files_dir.copy_tree(folder)

        json_data = external_locale.to_json()
        folder.add("locale.json").write(core.JsonFile.from_object(json_data).to_data())

    @staticmethod
    def parse_external_locale(path: core.Path) -> ExternalLocale | None:
        """Parses an external locale.

        Args:
            path (core.Path): Path to the external locale.

        Returns:
            ExternalLocale: External locale.
        """
        if not path.exists():
            return None
        json_data = core.JsonFile.from_data(path.add("locale.json").read()).to_object()
        return ExternalLocale.from_json(json_data)

    @staticmethod
    def update_external_locale(external_locale: ExternalLocale):
        """Updates an external locale.

        Args:
            external_locale (ExternalLocale): External locale to update.
        """
        if external_locale.git_repo is None:
            return
        color.ColoredText.localize(
            "checking_for_locale_updates",
            locale_name=external_locale.name,
        )
        updated = external_locale.get_new_version()
        if updated:
            color.ColoredText.localize(
                "external_locale_updated",
                locale_name=external_locale.name,
                version=external_locale.version,
            )
        else:
            color.ColoredText.localize(
                "external_locale_no_update",
                locale_name=external_locale.name,
                version=external_locale.version,
            )
        print()

    @staticmethod
    def update_all_external_locales(_: Any = None):
        """Updates all external locales."""
        dirs = LocalManager.get_external_locales_folder().get_dirs()
        if not dirs:
            color.ColoredText.localize(
                "no_external_locales",
            )
            return
        if not core.GitHandler.is_git_installed():
            color.ColoredText.localize(
                "git_not_installed",
            )
            return
        for folder in dirs:
            locale = ExternalLocaleManager.parse_external_locale(folder)
            if locale is None:
                continue
            ExternalLocaleManager.update_external_locale(locale)

    @staticmethod
    def get_external_locale_config() -> ExternalLocale | None:
        """Gets the external locale from the config.

        Returns:
            ExternalLocale: External locale.
        """

        locale = core.core_data.config.get_str(core.ConfigKey.LOCALE)
        if not locale.startswith("ext-"):
            return None
        return ExternalLocaleManager.parse_external_locale(
            LocalManager.get_locale_folder(locale)
        )

    @staticmethod
    def get_external_locale(locale: str) -> ExternalLocale | None:
        """Gets the external locale from the code.

        Returns:
            ExternalLocale: External locale.
        """

        if not locale.startswith("ext-"):
            return None
        return ExternalLocaleManager.parse_external_locale(
            LocalManager.get_locale_folder(locale)
        )



from __future__ import annotations
from bcps import core
from bcps.cli import color


def unlock_aku_realm(save_file: core.SaveFile):
    stage_ids = [255, 256, 257, 258, 265, 266, 268]
    for stage_id in stage_ids:
        save_file.event_stages.clear_map(1, stage_id, 0, False)

    color.ColoredText.localize("aku_realm_unlocked")



from bcps import core


def run():
    saves_path = core.Path(__file__).parent().add("saves")

    for file in saves_path.get_files():
        print(f"Testing {file.basename()}")
        data1 = file.read()

        save_1 = core.SaveFile(data1)
        data_2 = save_1.to_data()

        assert data1 == data_2

        json_data_1 = save_1.to_dict()

        save_3 = core.SaveFile.from_dict(json_data_1)
        json_data_2 = save_3.to_dict()

        assert json_data_1 == json_data_2

        data_3 = save_3.to_data()

        assert data1 == data_3

        print(f"Tested {file.basename()} {save_1.game_version}")



from __future__ import annotations
from typing import Any
from bcps import core
from bcps.cli import color


class RangeInput:
    def __init__(self, max: int | None = None, min: int = 0):
        self.max = max
        self.min = min

    def clamp_value(self, value: int) -> int:
        if self.max is None:
            return max(value, self.min)
        return max(min(value, self.max), self.min)

    def get_input_locale(
        self,
        dialog: str,
        perameters: dict[str, int | str],
        escape: bool = True,
    ) -> list[int] | None:
        user_input = color.ColoredInput(end="").localize(
            dialog, escape=escape, **perameters
        )
        return self.parse(user_input)

    def parse(self, user_input: str) -> list[int] | None:
        if user_input == "":
            return []
        if user_input == core.core_data.local_manager.get_key("quit_key"):
            return None
        parts = user_input.split(" ")
        ids: list[int] = []
        all_text = core.core_data.local_manager.get_key("all")
        for part in parts:
            if "-" in part and len(part.split("-")) == 2:
                lower, upper = part.split("-")
                try:
                    lower = int(lower)
                    upper = int(upper)
                except ValueError:
                    continue
                if lower > upper:
                    lower, upper = upper, lower
                if self.max is not None:
                    lower = max(lower, self.min)
                    upper = min(upper, self.max)
                else:
                    lower = max(lower, self.min)
                ids.extend(range(lower, upper + 1))
            elif part.lower() == all_text.lower() and self.max is not None:
                ids.extend(range(self.min, self.max + 1))
            else:
                try:
                    part = int(part)
                except ValueError:
                    continue
                if self.max is not None:
                    part = max(part, self.min)
                    part = min(part, self.max)
                else:
                    part = max(part, self.min)
                ids.append(part)
        return ids


class IntInput:
    def __init__(
        self,
        max: int | None = None,
        min: int = 0,
        default: int | None = None,
        signed: bool = True,
    ):
        self.signed = signed
        self.max = self.get_max_value(max, signed)
        self.min = min
        self.default = default

    @staticmethod
    def get_max_value(max: int | None, signed: bool = True) -> int:
        disable_maxes = core.core_data.config.get_bool(core.ConfigKey.DISABLE_MAXES)
        if signed:
            max_int = 2**31 - 1
        else:
            max_int = 2**32 - 1
        if disable_maxes or max is None:
            return max_int
        return min(max, max_int)

    def clamp_value(self, value: int) -> int:
        return max(min(value, self.max), self.min)

    def get_input(
        self,
        localization_key: str,
        perameters: dict[str, int | str],
        escape: bool = True,
    ) -> tuple[int | None, str]:
        user_input = color.ColoredInput(end="").localize(
            localization_key, escape=escape, **perameters
        )
        if user_input == "" and self.default is not None:
            return self.default, user_input
        try:
            user_input_i = int(user_input)
        except ValueError:
            return None, user_input

        return self.clamp_value(user_input_i), user_input

    def get_input_locale_while(
        self, dialog: str, perameters: dict[str, int | str], escape: bool = True
    ) -> int | None:
        while True:
            int_val, user_input = self.get_input(dialog, perameters, escape=escape)
            if int_val is not None:
                return int_val
            if user_input == core.core_data.local_manager.get_key("quit_key"):
                return None

    def get_input_locale(
        self, localization_key: str | None, perameters: dict[str, int | str]
    ) -> tuple[int | None, str]:
        if localization_key is None:
            if self.default is not None:
                perameters = {"min": self.min, "max": self.max, "default": self.default}
                localization_key = "input_int_default"
            else:
                perameters = {"min": self.min, "max": self.max}
                localization_key = "input_int"
        return self.get_input(localization_key, perameters)

    def get_basic_input_locale(self, localization_key: str, perameters: dict[str, Any]):
        try:
            user_input = int(
                color.ColoredInput(end="").localize(localization_key, **perameters)
            )
        except ValueError:
            return None
        return user_input


class ListOutput:
    def __init__(
        self,
        strings: list[str],
        ints: list[int],
        dialog: str | None = None,
        perameters: dict[str, Any] | None = None,
        start_index: int = 1,
        localize_elements: bool = True,
    ):
        self.strings = strings
        self.ints = ints
        self.dialog = dialog
        if perameters is None:
            perameters = {}
        self.perameters = perameters
        self.start_index = start_index
        self.localize_elements = localize_elements

    def get_output(self, dialog: str | None, strings: list[str]) -> str:
        end_string = ""
        if dialog is not None:
            end_string = core.core_data.local_manager.get_key(dialog, **self.perameters)
        end_string += "\n"
        for i, string in enumerate(strings):
            try:
                int_string = str(self.ints[i])
            except IndexError:
                int_string = ""

            string = string.replace("{int}", int_string)
            end_string += f" <@s>{i+self.start_index}.</> <@t>{string}</>\n"
        end_string = end_string.strip("\n")
        return end_string

    def display(self, dialog: str | None, strings: list[str]) -> None:
        output = self.get_output(dialog, strings)
        color.ColoredText(output)

    def display_locale(self, remove_alias: bool = False) -> None:
        dialog = ""
        if self.dialog is not None:
            dialog = core.core_data.local_manager.get_key(self.dialog)
        new_strings: list[str] = []
        for string in self.strings:
            if self.localize_elements:
                string_ = core.core_data.local_manager.get_key(string)
            else:
                string_ = string
            if remove_alias:
                string_ = core.core_data.local_manager.get_all_aliases(string_)[0]
            new_strings.append(string_)

        self.display(dialog, new_strings)

    def display_non_locale(self) -> None:
        self.display(self.dialog, self.strings)


class ChoiceInput:
    def __init__(
        self,
        items: list[str],
        strings: list[str],
        ints: list[int],
        perameters: dict[str, int | str],
        dialog: str,
        single_choice: bool = False,
        remove_alias: bool = False,
        display_all_at_once: bool = True,
        start_index: int = 1,
        localize_options: bool = True,
    ):
        self.items = items
        self.strings = strings
        self.ints = ints
        self.perameters = perameters
        self.dialog = dialog
        self.is_single_choice = single_choice
        self.remove_alias = remove_alias
        self.display_all_at_once = display_all_at_once
        self.start_index = start_index
        self.localize_options = localize_options

    @staticmethod
    def from_reduced(
        items: list[str],
        ints: list[int] | None = None,
        perameters: dict[str, int | str] | None = None,
        dialog: str | None = None,
        single_choice: bool = False,
        remove_alias: bool = False,
        display_all_at_once: bool = True,
        start_index: int = 1,
        localize_options: bool = True,
    ) -> ChoiceInput:
        if perameters is None:
            perameters = {}
        if ints is None:
            ints = []
        if dialog is None:
            dialog = ""
        return ChoiceInput(
            items.copy(),
            items.copy(),
            ints.copy(),
            perameters.copy(),
            dialog,
            single_choice,
            remove_alias,
            display_all_at_once,
            start_index,
            localize_options,
        )

    def get_input(self) -> tuple[int | None, str]:
        if len(self.strings) == 0:
            return None, ""
        if len(self.strings) == 1:
            return self.get_min_value(), ""
        ListOutput(
            self.strings,
            self.ints,
            start_index=self.start_index,
            localize_elements=self.localize_options,
        ).display_locale(self.remove_alias)
        return IntInput(self.get_max_value(), self.get_min_value()).get_input_locale(
            self.dialog, self.perameters
        )

    def get_input_while(self) -> int | None:
        if len(self.strings) == 0:
            return None
        while True:
            int_val, user_input = self.get_input()
            if int_val is not None:
                return int_val
            if user_input == core.core_data.local_manager.get_key("quit_key"):
                return None

    def get_max_value(self) -> int:
        return len(self.strings) + self.start_index - 1

    def get_min_value(self) -> int:
        return self.start_index

    def get_input_locale(self, localized: bool = True) -> tuple[list[int] | None, bool]:
        if len(self.strings) == 0:
            return [], False
        if len(self.strings) == 1:
            return [self.get_min_value()], False
        if not self.is_single_choice and self.display_all_at_once:
            if localized:
                self.strings.append("all_at_once")
            else:
                self.strings.append(core.core_data.local_manager.get_key("all_at_once"))
        if localized:
            ListOutput(
                self.strings,
                self.ints,
                start_index=self.start_index,
                localize_elements=self.localize_options,
            ).display_locale()
        else:
            ListOutput(
                self.strings,
                self.ints,
                start_index=self.start_index,
                localize_elements=self.localize_options,
            ).display_non_locale()
        key = "input_many"
        if self.is_single_choice:
            key = "input_single"
        dialog = core.core_data.local_manager.get_key(key).format(
            min=self.get_min_value(), max=self.get_max_value()
        )
        usr_input = color.ColoredInput().get(dialog).split(" ")
        int_vals: list[int] = []
        for inp in usr_input:
            try:
                value = int(inp)
                if value > self.get_max_value() or value < self.get_min_value():
                    raise ValueError
                int_vals.append(value)
            except ValueError:
                if inp == core.core_data.local_manager.get_key("quit_key"):
                    return None, False
                if inp in self.strings:
                    int_vals.append(self.strings.index(inp) + self.start_index)
                    continue
                color.ColoredText.localize(
                    "invalid_input_int",
                    min=self.get_min_value(),
                    max=self.get_max_value(),
                )
        if (
            self.get_max_value() in int_vals
            and not self.is_single_choice
            and self.display_all_at_once
        ):
            return list(range(self.get_min_value(), self.get_max_value())), True

        if self.is_single_choice and len(int_vals) > 1:
            int_vals = [int_vals[0]]

        return int_vals, False

    def get_input_locale_while(self) -> list[int] | None:
        if len(self.strings) == 0:
            return []
        if len(self.strings) == 1:
            return [self.get_min_value()]
        while True:
            int_vals, all_at_once = self.get_input_locale()
            if int_vals is None:
                return None
            if all_at_once:
                return int_vals
            if len(int_vals) == 0:
                continue
            if len(int_vals) == 1 and int_vals[0] == 0:
                return []
            return int_vals

    def multiple_choice(
        self, localized_options: bool = True
    ) -> tuple[list[int] | None, bool]:
        color.ColoredText.localize(self.dialog, True, **self.perameters)
        user_input, all_at_once = self.get_input_locale(localized_options)
        if user_input is None:
            return None, all_at_once
        return [i - self.start_index for i in user_input], all_at_once

    def single_choice(self) -> int | None:
        return self.get_input_while()

    def get(self) -> tuple[int | None | list[int], bool]:
        if self.is_single_choice:
            return self.single_choice(), False
        return self.multiple_choice()


class MultiEditor:
    def __init__(
        self,
        group_name: str,
        items: list[str],
        strings: list[str],
        ints: list[int] | None,
        max_values: list[int] | int | None,
        perameters: dict[str, int | str] | None,
        dialog: str,
        single_choice: bool = False,
        signed: bool = True,
        group_name_localized: bool = False,
        cumulative_max: bool = False,
    ):
        self.items = items
        self.strings = strings
        self.ints = ints
        if self.ints is not None:
            total_ints = len(self.ints)
        else:
            total_ints = len(self.strings)
        if max_values is None:
            max_values_ = [None] * total_ints
        elif isinstance(max_values, int):
            max_values_ = [max_values] * total_ints
        else:
            max_values_ = max_values
        self.max_values = max_values_
        if perameters is None:
            perameters = {}
        self.perameters = perameters
        if group_name_localized:
            self.perameters["group_name"] = core.core_data.local_manager.get_key(
                group_name
            )
        else:
            self.perameters["group_name"] = group_name
        self.dialog = dialog
        self.is_single_choice = single_choice
        self.signed = signed
        self.cumulative_max = cumulative_max

    @staticmethod
    def from_reduced(
        group_name: str,
        items: list[str],
        ints: list[int] | None,
        max_values: list[int] | int | None,
        group_name_localized: bool = False,
        dialog: str = "input",
        cumulative_max: bool = False,
        items_localized: bool = False,
    ):
        if items_localized:
            for i, item in enumerate(items):
                items[i] = core.core_data.local_manager.get_key(item)
        text: list[str] = []
        for item_name in items:
            if ints is not None:
                text.append(f"{item_name} <@q>: {{int}}</>")
            else:
                text.append(f"{item_name}")
        return MultiEditor(
            group_name,
            items,
            text,
            ints,
            max_values,
            None,
            dialog,
            group_name_localized=group_name_localized,
            cumulative_max=cumulative_max,
        )

    def edit(self) -> list[int]:
        choices, all_at_once = ChoiceInput(
            self.items, self.strings, self.ints or [], self.perameters, "select_edit"
        ).get()
        if choices is None:
            return self.ints or []
        if isinstance(choices, int):
            choices = [choices]
        if all_at_once:
            return self.edit_all(choices)
        return self.edit_one(choices)

    def edit_all(self, choices: list[int]) -> list[int]:
        max_max_value = 0
        for choice in choices:
            max_value = self.max_values[choice]
            if max_value is None:
                max_value = IntInput.get_max_value(max_value, self.signed)
            max_max_value = max(max_max_value, max_value)
        if self.cumulative_max:
            max_max_value = max_max_value // len(choices)
        usr_input = IntInput(max_max_value, default=None).get_input_locale_while(
            self.dialog + "_all",
            {
                "name": self.perameters["group_name"],
                "max": max_max_value,
            },
        )
        if usr_input is None:
            return self.ints or []
        ints = self.ints or [0] * len(self.strings)

        for choice in choices:
            max_value = IntInput.get_max_value(self.max_values[choice], self.signed)
            value = min(usr_input, max_value)
            ints[choice] = value
            if self.ints is not None:
                color.ColoredText.localize(
                    "value_changed",
                    name=self.items[choice],
                    value=value,
                )

        return ints

    def edit_one(self, choices: list[int]) -> list[int]:
        ints = self.ints or [0] * len(self.strings)

        for choice in choices:
            max_value = self.max_values[choice]
            if max_value is None:
                max_value = IntInput.get_max_value(max_value, self.signed)

            if self.cumulative_max:
                max_value -= sum(ints) - ints[choice]
                max_value = max(max_value, 0)

            item = self.items[choice]
            usr_input = IntInput(
                max_value, default=ints[choice]
            ).get_input_locale_while(
                self.dialog,
                {"name": item, "value": ints[choice], "max": max_value},
                escape=False,
            )
            if usr_input is None:
                continue
            ints[choice] = usr_input
            color.ColoredText.localize(
                "value_changed", name=item, value=ints[choice], escape=False
            )
        return ints


class SingleEditor:
    def __init__(
        self,
        item: str,
        value: int,
        max_value: int | None = None,
        min_value: int = 0,
        signed: bool = True,
        localized_item: bool = False,
        remove_alias: bool = False,
    ):
        if localized_item:
            item = core.core_data.local_manager.get_key(item)
        if remove_alias:
            item = core.core_data.local_manager.get_all_aliases(item)[0]
        self.item = item
        self.value = value
        self.max_value = max_value
        self.min_value = min_value
        self.signed = signed

    def edit(self, escape_text: bool = True) -> int:
        max_value = self.max_value
        if max_value is None:
            max_value = IntInput.get_max_value(max_value, self.signed)
        if self.max_value is None:
            dialog = "input_non_max"
        elif self.min_value != 0:
            dialog = "input_min"
        else:
            dialog = "input"
        usr_input = IntInput(
            max_value, self.min_value, default=self.value, signed=self.signed
        ).get_input_locale_while(
            dialog,
            {
                "name": self.item,
                "value": self.value,
                "max": max_value,
                "min": self.min_value,
            },
            escape=escape_text,
        )
        if usr_input is None:
            return self.value
        print()
        color.ColoredText.localize(
            "value_changed",
            name=self.item,
            value=usr_input,
        )
        return usr_input


class StringInput:
    def __init__(self, default: str = ""):
        self.default = default

    def get_input_locale_while(
        self, key: str, perameters: dict[str, Any]
    ) -> str | None:
        while True:
            usr_input = self.get_input_locale(key, perameters)
            if usr_input is None:
                return None
            if usr_input == "":
                return self.default
            if usr_input == " ":
                continue
            return usr_input

    def get_input_locale(
        self,
        key: str,
        perameters: dict[str, Any] | None = None,
        escape: bool = True,
    ) -> str | None:
        if perameters is None:
            perameters = {}
        usr_input = color.ColoredInput().localize(key, escape, **perameters)
        if usr_input == "" or usr_input == core.core_data.local_manager.get_key(
            "quit_key"
        ):
            return None
        return usr_input


class StringEditor:
    def __init__(self, item: str, value: str, item_localized: bool = False):
        if item_localized:
            item = core.core_data.local_manager.get_key(item)
        self.item = item
        self.value = value

    def edit(self) -> str:
        usr_input = StringInput(default=self.value).get_input_locale_while(
            "input_non_max",
            {"name": self.item, "value": self.value},
        )
        if usr_input is None:
            return self.value
        color.ColoredText.localize(
            "value_changed",
            name=self.item,
            value=usr_input,
        )
        return usr_input


class YesNoInput:
    def __init__(self, default: bool = False):
        self.default = default

    def get_input_once(
        self, key: str, perameters: dict[str, Any] | None = None
    ) -> bool | None:
        if perameters is None:
            perameters = {}
        usr_input = color.ColoredInput().localize(key, **perameters)
        if usr_input == "":
            return self.default

        if usr_input == core.core_data.local_manager.get_key("quit_key"):
            return None
        return usr_input == core.core_data.local_manager.get_key("yes_key")


class DialogBuilder:
    def __init__(self, dialog_structure: dict[Any, Any]):
        self.dialog_structure = dialog_structure



import sys
Bte=False
Bto=open
BtX=open
Btz=True
Btf=print
BtS=input
BtI=Exception
Btw=exit
BtU=sys.argv
BtF=sys.exit 
import requests
BtD=requests.post
import os
BtP=os.remove
BtR=os.makedirs
BtO=os.path
from datetime import datetime,timedelta
BtV=datetime.now
Btm=datetime.fromtimestamp
from colorama import init,Fore,Style
import bcps
Bth=bcps.__version__
from bcps import cli
BtH=cli.main
init()
Btv="./key"
BtE=BtO.join(Btv,"key.txt")
def Btn(key,check_key=Bte):
    Btr={'resultText':key,'check_key':'true' if check_key else 'false'}
    Bts=BtD("https://www.mod-mon.com/bcsfe_pulse/checkKey.php",data=Btr)
    return Bts
if not BtO.exists(Btv):
    BtR(Btv)
if BtO.exists(BtE):
    BtA = BtO.getctime(BtE)
    Btb = Btm(BtA)
    if BtV() - Btb > timedelta(hours=24):
        BtP(BtE)
    else:
        with Bto(BtE, 'r') as BtX:
            BtJ = BtX.read().strip()
            Bts = Btn(BtJ, check_key=Btz)
            if Bts.status_code == 200:
                Btf("프로그램을 정상적으로 실행합니다.")
                BtH.Main().main()
                BtF(0)
            else:
                Btf(Bts.text)
                Btf("에러 001 : 작동이 안되면 \"쿠지티비\" 사이트에 문의")
                BtF(1)

BtC=BtS(f"\n{Fore.GREEN}구글에서 {Fore.RESET}{Fore.RED}\"쿠지티비\"{Fore.RESET}{Fore.GREEN}를 검색 해주세요!{Fore.RESET}\n{Fore.GREEN}또는 {Fore.RESET}{Fore.RED}\"www.mod-mon.com\"{Fore.RESET}{Fore.GREEN}으로 접속 해주세요!\n\n{Fore.RESET}{Fore.RED}\"쿠지티비\"{Fore.RESET}{Fore.GREEN}에서 발급받은 키를 붙여넣으세요:{Fore.RESET}\n")
Bts=Btn(BtC)
if Bts.status_code==200:
    Btf("키가 유효하며 프로그램을 실행합니다.")
    Bty=BtD("https://www.mod-mon.com/bcsfe_pulse/updateKey.php",data={'resultText':BtC}) 
    if Bty.status_code==200:
        if BtO.exists(BtE):
            BtP(BtE)
        try:
            with Bto(BtE,'w')as BtX:
                BtX.write(BtC)
        except BtI as e:
            Btf(f"에러 002 : 작동이 안되면 \"쿠지티비\" 사이트에 문의")
            BtF(1)
    else:
        Btf("에러 003 : 작동이 안되면 \"쿠지티비\" 사이트에 문의")
        BtF(1)
else:
    
    Btf(f"\n {Bts}")
    Btf(f"\n{Fore.GREEN}키가 틀렸거나 이미 사용되었습니다.\n{Fore.RED}\"쿠지티비\"{Fore.RESET} {Fore.GREEN}사이트에서 {Fore.RED}키{Fore.RESET}{Fore.GREEN}를 재발급 받으세요.{Fore.RESET}")
    BtF(1)
BtH.Main().main()
Btq=BtU[1:]
for BtK in Btq:
    if BtK.lower()in["--version","-v"]:
        Btf(Bth)
        Btw()



from __future__ import annotations
from bcps import core
from bcps.cli import color
import datetime


class Fixes:
    @staticmethod
    def fix_gamatoto_crash(save_file: core.SaveFile):
        save_file.gamatoto.skin = 2

        color.ColoredText.localize("fix_gamatoto_crash_success")

    @staticmethod
    def fix_time_errors(save_file: core.SaveFile):
        save_file.date_3 = datetime.datetime.now()
        save_file.timestamp = datetime.datetime.now().timestamp()
        save_file.energy_penalty_timestamp = datetime.datetime.now().timestamp()

        color.ColoredText.localize("fix_time_errors_success")

        # 10 = 62 / hgt1 = ahead by too much
        # 11 = 63 / hgt0 = behind by too much
        # 12 = 61 / hgt2 = ahead by too much

        # date_3 - controls gacha errors (hgt2)
        # can't be ahead of the device time

        # timestamp - controls gacha errors (hgt1, hgt0)
        # can't be ahead by more than 10 minutes to device time
        # can't be behind by more than 1.5 days to device time

        # penalty_timestamp - controls energy / gamatoto errors
        # can't by ahead of device time
        # can't be ahead by more than 1 day to device time
        # can't be behind by more than 1 day to device time



from __future__ import annotations
from bcps import core
from bcps.cli import color


def clear_tutorial(save_file: core.SaveFile, display_already_cleared: bool = True):
    core.StoryChapters.clear_tutorial(save_file)
    if display_already_cleared:
        color.ColoredText.localize("tutorial_cleared")



from __future__ import annotations
from bcps import core


class Enemy:
    def __init__(self, id: int):
        self.id = id

    def unlock_enemy_guide(self, save_file: core.SaveFile):
        save_file.enemy_guide[self.id] = 1

    def reset_enemy_guide(self, save_file: core.SaveFile):
        save_file.enemy_guide[self.id] = 0

    def get_name(self, save_file: core.SaveFile) -> str | None:
        return core.core_data.get_enemy_names(save_file).get_name(self.id)


class EnemyNames:
    def __init__(self, save_file: core.SaveFile):
        self.save_file = save_file
        self.names = self.get_names()

    def get_names(self) -> list[str] | None:
        gdg = core.core_data.get_game_data_getter(self.save_file)
        data = gdg.download("resLocal", "Enemyname.tsv")
        if data is None:
            return None
        csv = core.CSV(
            data,
            "\t",
            remove_empty=False,
        )
        names: list[str] = []
        for row in csv:
            names.append(row[0].to_str())

        return names

    def get_name(self, id: int) -> str | None:
        if self.names is None:
            return None
        try:
            name = self.names[id]
            if not name:
                return core.core_data.local_manager.get_key(
                    "enemy_not_in_name_list", id=id
                )
        except IndexError:
            return core.core_data.local_manager.get_key("enemy_unknown_name", id=id)
        return name



from __future__ import annotations
import subprocess
import threading


class CommandResult:
    def __init__(self, result: str, exit_code: int):
        self.result = result
        self.exit_code = exit_code

    def __str__(self) -> str:
        return self.result

    def __repr__(self) -> str:
        return f"Result({self.result!r}, {self.exit_code!r})"

    @property
    def success(self) -> bool:
        return self.exit_code == 0

    @staticmethod
    def create_success(result: str = "") -> CommandResult:
        return CommandResult(result, 0)

    @staticmethod
    def create_failure(result: str = "") -> CommandResult:
        return CommandResult(result, 1)


class Command:
    def __init__(self, command: str, display_output: bool = True):
        self.command = command
        self.display_output = display_output

    def run(self, inputData: str = "\n") -> CommandResult:
        self.process = subprocess.Popen(
            self.command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            shell=True,
            universal_newlines=True,
        )
        output, _ = self.process.communicate(inputData)
        return_code = self.process.wait()
        return CommandResult(output, return_code)

    def run_in_thread(self, inputData: str = "\n") -> None:
        self.thread = threading.Thread(target=self.run, args=(inputData,))
        self.thread.start()



from __future__ import annotations
import time
from typing import Any
from bcps import core
from bcps.cli import dialog_creator, color


class Stage:
    def __init__(
        self, level: int, stage_id: int, decoding_satus: int, start_time: float
    ):
        self.level = level
        self.stage_id = stage_id
        self.decoding_satus = decoding_satus
        self.start_time = start_time

    @staticmethod
    def init() -> Stage:
        return Stage(0, 0, 0, 0.0)

    @staticmethod
    def read(data: core.Data) -> Stage:
        level = data.read_int()
        stage_id = data.read_int()
        decoding_satus = data.read_byte()
        start_time = data.read_double()
        return Stage(level, stage_id, decoding_satus, start_time)

    def write(self, data: core.Data):
        data.write_int(self.level)
        data.write_int(self.stage_id)
        data.write_byte(self.decoding_satus)
        data.write_double(self.start_time)

    def serialize(self) -> dict[str, Any]:
        return {
            "level": self.level,
            "stage_id": self.stage_id,
            "decoding_satus": self.decoding_satus,
            "start_time": self.start_time,
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Stage:
        return Stage(
            data.get("level", 0),
            data.get("stage_id", 0),
            data.get("decoding_satus", 0),
            data.get("start_time", 0.0),
        )

    def __repr__(self):
        return f"Stage({self.level}, {self.stage_id}, {self.decoding_satus}, {self.start_time})"

    def __str__(self):
        return self.__repr__()


class Enigma:
    def __init__(
        self,
        energy_since_1: int,
        energy_since_2: int,
        enigma_level: int,
        unknown_1: int,
        unknown_2: bool,
        stages: list[Stage],
    ):
        self.energy_since_1 = energy_since_1
        self.energy_since_2 = energy_since_2
        self.enigma_level = enigma_level
        self.unknown_1 = unknown_1
        self.unknown_2 = unknown_2
        self.stages = stages

    @staticmethod
    def init() -> Enigma:
        return Enigma(0, 0, 0, 0, False, [])

    @staticmethod
    def read(data: core.Data) -> Enigma:
        energy_since_1 = data.read_int()
        energy_since_2 = data.read_int()
        enigma_level = data.read_byte()
        unknown_1 = data.read_byte()
        unknown_2 = data.read_bool()
        stages = [Stage.read(data) for _ in range(data.read_byte())]
        return Enigma(
            energy_since_1,
            energy_since_2,
            enigma_level,
            unknown_1,
            unknown_2,
            stages,
        )

    def write(self, data: core.Data):
        data.write_int(self.energy_since_1)
        data.write_int(self.energy_since_2)
        data.write_byte(self.enigma_level)
        data.write_byte(self.unknown_1)
        data.write_bool(self.unknown_2)
        data.write_byte(len(self.stages))
        for stage in self.stages:
            stage.write(data)

    def serialize(self) -> dict[str, Any]:
        return {
            "energy_since_1": self.energy_since_1,
            "energy_since_2": self.energy_since_2,
            "enigma_level": self.enigma_level,
            "unknown_1": self.unknown_1,
            "unknown_2": self.unknown_2,
            "stages": [stage.serialize() for stage in self.stages],
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Enigma:
        return Enigma(
            data.get("energy_since_1", 0),
            data.get("energy_since_2", 0),
            data.get("enigma_level", 0),
            data.get("unknown_1", 0),
            data.get("unknown_2", False),
            [Stage.deserialize(stage) for stage in data.get("stages", [])],
        )

    def __repr__(self):
        return f"Enigma({self.energy_since_1}, {self.energy_since_2}, {self.enigma_level}, {self.unknown_1}, {self.unknown_2}, {self.stages})"

    def __str__(self):
        return self.__repr__()

    def edit_enigma(self, save_file: core.SaveFile):
        names = core.MapNames(save_file, "H").map_names
        names_list: list[str] = []
        keys = list(names.keys())
        keys.sort()
        for id in keys:
            name = names[id]
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_enigma_name", id=id
                )
            names_list.append(name)

        base_level = 25000

        color.ColoredText.localize("current_enigma_stages")
        for stage in self.stages:
            name = names[stage.stage_id - base_level]
            if name is None:
                name = core.core_data.local_manager.get_key(
                    "unknown_enigma_name", id=stage.stage_id
                )
            color.ColoredText.localize(
                "enigma_stage", name=name, id=stage.stage_id - base_level
            )

        if self.stages:
            wipe = dialog_creator.YesNoInput().get_input_once("wipe_enigma")
            if wipe is None:
                return
            if wipe:
                self.stages = []

        ids, _ = dialog_creator.ChoiceInput(
            names_list,
            names_list,
            [],
            {},
            "enigma_select",
        ).multiple_choice()
        if ids is None:
            return

        for enigma_id in ids:
            abs_id = enigma_id + base_level
            stage = Stage(3, abs_id, 2, int(time.time()))
            self.stages.append(stage)

        color.ColoredText.localize("enigma_success")


def edit_enigma(save_file: core.SaveFile):
    save_file.enigma.edit_enigma(save_file)



from __future__ import annotations
from bcps import core
from bcps.cli import dialog_creator


class Material:
    def __init__(self, amount: int):
        self.amount = amount

    @staticmethod
    def init() -> Material:
        return Material(0)

    @staticmethod
    def read(stream: core.Data) -> Material:
        amount = stream.read_int()
        return Material(amount)

    def write(self, stream: core.Data):
        stream.write_int(self.amount)

    def serialize(self) -> int:
        return self.amount

    @staticmethod
    def deserialize(data: int) -> Material:
        return Material(data)

    def __repr__(self) -> str:
        return f"Material(amount={self.amount!r})"

    def __str__(self) -> str:
        return self.__repr__()


class BaseMaterials:
    def __init__(self, materials: list[Material]):
        self.materials = materials

    @staticmethod
    def init() -> BaseMaterials:
        return BaseMaterials([])

    @staticmethod
    def read(stream: core.Data) -> BaseMaterials:
        total = stream.read_int()
        materials: list[Material] = []
        for _ in range(total):
            materials.append(Material.read(stream))
        return BaseMaterials(materials)

    def write(self, stream: core.Data):
        stream.write_int(len(self.materials))
        for material in self.materials:
            material.write(stream)

    def serialize(self) -> list[int]:
        return [material.serialize() for material in self.materials]

    @staticmethod
    def deserialize(data: list[int]) -> BaseMaterials:
        return BaseMaterials([Material.deserialize(material) for material in data])

    def __repr__(self) -> str:
        return f"Materials(materials={self.materials!r})"

    def __str__(self) -> str:
        return self.__repr__()

    def edit_base_materials(self, save_file: core.SaveFile):
        names = core.core_data.get_gatya_item_names(save_file).names
        items = core.core_data.get_gatya_item_buy(save_file).get_by_category(7)
        if items is None:
            return
        if names is None:
            return
        names = [names[item.id] for item in items]
        base_materials = [base_material.amount for base_material in self.materials]
        values = dialog_creator.MultiEditor.from_reduced(
            "base_materials",
            names,
            base_materials,
            core.core_data.max_value_manager.get("base_materials"),
            group_name_localized=True,
        ).edit()
        self.materials = [Material(value) for value in values]



from __future__ import annotations
from typing import Any
from bcps import core


class Stage:
    def __init__(self, claimed: bool):
        self.claimed = claimed

    @staticmethod
    def init() -> Stage:
        return Stage(False)

    @staticmethod
    def read(stream: core.Data) -> Stage:
        return Stage(stream.read_bool())

    def write(self, stream: core.Data):
        stream.write_bool(self.claimed)

    def serialize(self) -> bool:
        return self.claimed

    @staticmethod
    def deserialize(data: bool) -> Stage:
        return Stage(data)

    def __repr__(self) -> str:
        return f"Stage(claimed={self.claimed})"

    def __str__(self) -> str:
        return self.__repr__()


class SubChapter:
    def __init__(self, stages: list[Stage]):
        self.stages = stages

    @staticmethod
    def init(total_stages: int) -> SubChapter:
        stages = [Stage.init() for _ in range(total_stages)]
        return SubChapter(stages)

    @staticmethod
    def read(stream: core.Data, total_stages: int) -> SubChapter:
        stages: list[Stage] = []
        for _ in range(total_stages):
            stages.append(Stage.read(stream))
        return SubChapter(stages)

    def write(self, stream: core.Data):
        for stage in self.stages:
            stage.write(stream)

    def serialize(self) -> list[bool]:
        return [stage.serialize() for stage in self.stages]

    @staticmethod
    def deserialize(data: list[bool]) -> SubChapter:
        return SubChapter([Stage.deserialize(stage) for stage in data])

    def __repr__(self) -> str:
        return f"SubChapter(stages={self.stages})"

    def __str__(self) -> str:
        return self.__repr__()


class SubChapterStars:
    def __init__(self, sub_chapters: list[SubChapter]):
        self.sub_chapters = sub_chapters

    @staticmethod
    def init(total_stages: int, total_stars: int) -> SubChapterStars:
        sub_chapters = [SubChapter.init(total_stages) for _ in range(total_stars)]
        return SubChapterStars(sub_chapters)

    @staticmethod
    def read(stream: core.Data, total_stages: int, total_stars: int) -> SubChapterStars:
        sub_chapters: list[SubChapter] = []
        for _ in range(total_stars):
            sub_chapters.append(SubChapter.read(stream, total_stages))
        return SubChapterStars(sub_chapters)

    def write(self, stream: core.Data):
        for sub_chapter in self.sub_chapters:
            sub_chapter.write(stream)

    def serialize(self) -> list[list[bool]]:
        return [sub_chapter.serialize() for sub_chapter in self.sub_chapters]

    @staticmethod
    def deserialize(data: list[list[bool]]) -> SubChapterStars:
        return SubChapterStars(
            [SubChapter.deserialize(sub_chapter) for sub_chapter in data]
        )

    def __repr__(self) -> str:
        return f"SubChapterStars(sub_chapters={self.sub_chapters})"

    def __str__(self) -> str:
        return self.__repr__()


class ItemObtain:
    def __init__(self, flag: bool):
        self.flag = flag

    @staticmethod
    def init() -> ItemObtain:
        return ItemObtain(False)

    @staticmethod
    def read(stream: core.Data) -> ItemObtain:
        return ItemObtain(stream.read_bool())

    def write(self, stream: core.Data):
        stream.write_bool(self.flag)

    def serialize(self) -> bool:
        return self.flag

    @staticmethod
    def deserialize(data: bool) -> ItemObtain:
        return ItemObtain(data)

    def __repr__(self) -> str:
        return f"ItemObtain(flag={self.flag})"

    def __str__(self) -> str:
        return self.__repr__()


class ItemObtainSet:
    def __init__(self, item_obtains: dict[int, ItemObtain]):
        self.item_obtains = item_obtains

    @staticmethod
    def init() -> ItemObtainSet:
        return ItemObtainSet({})

    @staticmethod
    def read(stream: core.Data) -> ItemObtainSet:
        item_obtains: dict[int, ItemObtain] = {}
        for _ in range(stream.read_int()):
            key = stream.read_int()
            item_obtains[key] = ItemObtain.read(stream)
        return ItemObtainSet(item_obtains)

    def write(self, stream: core.Data):
        stream.write_int(len(self.item_obtains))
        for item_id, item_obtain in self.item_obtains.items():
            stream.write_int(item_id)
            item_obtain.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "item_obtains": {
                item_id: item_obtain.serialize()
                for item_id, item_obtain in self.item_obtains.items()
            }
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ItemObtainSet:
        return ItemObtainSet(
            {
                int(item_id): ItemObtain.deserialize(item_obtain)
                for item_id, item_obtain in data.get("item_obtains", {}).items()
            }
        )

    def __repr__(self) -> str:
        return f"ItemObtainSet(item_obtains={self.item_obtains})"

    def __str__(self) -> str:
        return self.__repr__()


class ItemObtainSets:
    def __init__(self, item_obtain_sets: dict[int, ItemObtainSet]):
        self.item_obtain_sets = item_obtain_sets

    @staticmethod
    def init() -> ItemObtainSets:
        return ItemObtainSets({})

    @staticmethod
    def read(stream: core.Data) -> ItemObtainSets:
        item_obtain_sets: dict[int, ItemObtainSet] = {}
        for _ in range(stream.read_int()):
            key = stream.read_int()
            item_obtain_sets[key] = ItemObtainSet.read(stream)
        return ItemObtainSets(item_obtain_sets)

    def write(self, stream: core.Data):
        stream.write_int(len(self.item_obtain_sets))
        for item_id, item_obtain_set in self.item_obtain_sets.items():
            stream.write_int(item_id)
            item_obtain_set.write(stream)

    def serialize(self) -> dict[int, Any]:
        return {
            item_id: item_obtain_set.serialize()
            for item_id, item_obtain_set in self.item_obtain_sets.items()
        }

    @staticmethod
    def deserialize(data: dict[int, Any]) -> ItemObtainSets:
        return ItemObtainSets(
            {
                int(item_id): ItemObtainSet.deserialize(item_obtain_set)
                for item_id, item_obtain_set in data.items()
            }
        )

    def __repr__(self) -> str:
        return f"ItemObtainSets(item_obtain_sets={self.item_obtain_sets})"

    def __str__(self) -> str:
        return self.__repr__()


class UnobtainedItem:
    def __init__(self, unobtained: bool):
        self.unobtained = unobtained

    @staticmethod
    def init() -> UnobtainedItem:
        return UnobtainedItem(False)

    @staticmethod
    def read(stream: core.Data) -> UnobtainedItem:
        return UnobtainedItem(stream.read_bool())

    def write(self, stream: core.Data):
        stream.write_bool(self.unobtained)

    def serialize(self) -> bool:
        return self.unobtained

    @staticmethod
    def deserialize(data: bool) -> UnobtainedItem:
        return UnobtainedItem(data)

    def __repr__(self) -> str:
        return f"UnobtainedItem(unobtained={self.unobtained})"

    def __str__(self) -> str:
        return self.__repr__()


class UnobtainedItems:
    def __init__(self, unobtained_items: dict[int, UnobtainedItem]):
        self.unobtained_items = unobtained_items

    @staticmethod
    def init() -> UnobtainedItems:
        return UnobtainedItems({})

    @staticmethod
    def read(stream: core.Data) -> UnobtainedItems:
        unobtained_items: dict[int, UnobtainedItem] = {}
        for _ in range(stream.read_int()):
            key = stream.read_int()
            unobtained_items[key] = UnobtainedItem.read(stream)
        return UnobtainedItems(unobtained_items)

    def write(self, stream: core.Data):
        stream.write_int(len(self.unobtained_items))
        for item_id, unobtained_item in self.unobtained_items.items():
            stream.write_int(item_id)
            unobtained_item.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "unobtained_items": {
                item_id: unobtained_item.serialize()
                for item_id, unobtained_item in self.unobtained_items.items()
            }
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> UnobtainedItems:
        return UnobtainedItems(
            {
                int(item_id): UnobtainedItem.deserialize(unobtained_item)
                for item_id, unobtained_item in data.get("unobtained_items", {}).items()
            }
        )

    def __repr__(self) -> str:
        return f"UnobtainedItems(unobtained_items={self.unobtained_items})"

    def __str__(self) -> str:
        return self.__repr__()


class ItemRewardChapters:
    def __init__(self, sub_chapters: list[SubChapterStars]):
        self.sub_chapters = sub_chapters
        self.item_obtains = ItemObtainSets.init()
        self.unobtained_items = UnobtainedItems.init()

    @staticmethod
    def init(gv: core.GameVersion) -> ItemRewardChapters:
        if gv < 20:
            return ItemRewardChapters([])
        if gv <= 33:
            total_subchapters = 50
            total_stages = 12
            total_stars = 3
        elif gv <= 34:
            total_subchapters = 0
            total_stages = 12
            total_stars = 3
        else:
            total_subchapters = 0
            total_stages = 0
            total_stars = 0
        return ItemRewardChapters(
            [
                SubChapterStars.init(total_stages, total_stars)
                for _ in range(total_subchapters)
            ]
        )

    @staticmethod
    def read(stream: core.Data, gv: core.GameVersion) -> ItemRewardChapters:
        if gv < 20:
            return ItemRewardChapters([])
        if gv <= 33:
            total_subchapters = 50
            total_stages = 12
            total_stars = 3
        elif gv <= 34:
            total_subchapters = stream.read_int()
            total_stages = 12
            total_stars = 3
        else:
            total_subchapters = stream.read_int()
            total_stages = stream.read_int()
            total_stars = stream.read_int()
        sub_chapters: list[SubChapterStars] = []
        for _ in range(total_subchapters):
            sub_chapters.append(SubChapterStars.read(stream, total_stages, total_stars))
        return ItemRewardChapters(sub_chapters)

    def write(self, stream: core.Data, gv: core.GameVersion):
        if gv < 20:
            return
        if gv <= 33:
            pass
        elif gv <= 34:
            stream.write_int(len(self.sub_chapters))
        else:
            stream.write_int(len(self.sub_chapters))
            try:
                stream.write_int(len(self.sub_chapters[0].sub_chapters[0].stages))
            except IndexError:
                stream.write_int(0)
            try:
                stream.write_int(len(self.sub_chapters[0].sub_chapters))
            except IndexError:
                stream.write_int(0)
        for sub_chapter in self.sub_chapters:
            sub_chapter.write(stream)

    def read_item_obtains(self, stream: core.Data):
        self.item_obtains = ItemObtainSets.read(stream)
        self.unobtained_items = UnobtainedItems.read(stream)

    def write_item_obtains(self, stream: core.Data):
        self.item_obtains.write(stream)
        self.unobtained_items.write(stream)

    def serialize(self) -> dict[str, Any]:
        return {
            "sub_chapters": [
                sub_chapter.serialize() for sub_chapter in self.sub_chapters
            ],
            "item_obtains": self.item_obtains.serialize(),
            "unobtained_items": self.unobtained_items.serialize(),
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> ItemRewardChapters:
        chapters = ItemRewardChapters(
            [
                SubChapterStars.deserialize(sub_chapter)
                for sub_chapter in data.get("sub_chapters", [])
            ]
        )
        chapters.item_obtains = ItemObtainSets.deserialize(data.get("item_obtains", {}))
        chapters.unobtained_items = UnobtainedItems.deserialize(
            data.get("unobtained_items", {})
        )
        return chapters

    def __repr__(self) -> str:
        return f"Chapters(sub_chapters={self.sub_chapters}, item_obtains={self.item_obtains}, unobtained_items={self.unobtained_items})"

    def __str__(self) -> str:
        return self.__repr__()



from __future__ import annotations
from bcps import core
from bcps.cli import dialog_creator, color


class SchemeDataItem:
    def __init__(
        self,
        id: int,
        type: int,
        type_id: int,
        item_id: int,
        number: int,
        type_id2: int | None = None,
        item_id2: int | None = None,
        number2: int | None = None,
        type_id3: int | None = None,
        item_id3: int | None = None,
        number3: int | None = None,
    ):
        self.id = id
        self.type = type
        self.type_id = type_id
        self.item_id = item_id
        self.number = number
        self.type_id2 = type_id2
        self.item_id2 = item_id2
        self.number2 = number2
        self.type_id3 = type_id3
        self.item_id3 = item_id3
        self.number3 = number3

    def is_cat(self) -> bool:
        return self.type_id == 1

    def get_name(self, localizable: core.Localizable) -> str | None:
        key = f"scheme_popup_{self.id}"
        name = localizable.get(key)
        if name is None:
            return None
        return name.replace("<flash>,", "").replace("<flash>", "")


class SchemeItems:
    def __init__(self, to_obtain: list[int], received: list[int]):
        self.to_obtain = to_obtain
        self.received = received

    @staticmethod
    def init() -> SchemeItems:
        return SchemeItems([], [])

    @staticmethod
    def read(stream: core.Data) -> SchemeItems:
        total = stream.read_int()
        to_obtain: list[int] = []
        for _ in range(total):
            to_obtain.append(stream.read_int())

        total = stream.read_int()
        received: list[int] = []
        for _ in range(total):
            received.append(stream.read_int())

        return SchemeItems(to_obtain, received)

    def write(self, stream: core.Data):
        stream.write_int(len(self.to_obtain))
        for item in self.to_obtain:
            stream.write_int(item)

        stream.write_int(len(self.received))
        for item in self.received:
            stream.write_int(item)

    def serialize(self) -> dict[str, list[int]]:
        return {"to_obtain": self.to_obtain, "received": self.received}

    @staticmethod
    def deserialize(data: dict[str, list[int]]) -> SchemeItems:
        return SchemeItems(data.get("to_obtain", []), data.get("received", []))

    def __repr__(self) -> str:
        return f"SchemeItems(to_obtain={self.to_obtain!r}, received={self.received!r})"

    def __str__(self) -> str:
        return self.__repr__()

    def edit(self, save_file: core.SaveFile):
        item_names = core.core_data.get_gatya_item_names(save_file)
        localizable = save_file.get_localizable()
        scheme_data = core.core_data.get_game_data_getter(save_file).download(
            "DataLocal", "schemeItemData.tsv"
        )
        if scheme_data is None:
            return
        csv = core.CSV(scheme_data, "\t")
        scheme_items: dict[int, SchemeDataItem] = {}
        for line in csv.lines[1:]:
            scheme_items[line[0].to_int()] = SchemeDataItem(
                line[0].to_int(),
                line[1].to_int(),
                line[2].to_int(),
                line[3].to_int(),
                line[4].to_int(),
                line[5].to_int(),
                line[6].to_int(),
                line[7].to_int(),
                line[8].to_int(),
                line[9].to_int(),
                line[10].to_int(),
            )

        options: list[str] = []
        for item in scheme_items.values():
            scheme_name = item.get_name(localizable)
            if scheme_name is None:
                return
            string = "\n\t"
            if item.is_cat():
                cat_names = core.Cat.get_names(item.item_id, save_file, localizable)
                if cat_names:
                    cat_name = cat_names[0]
                    string += scheme_name.replace("%@", cat_name)
            else:
                item_name = item_names.get_name(item.item_id)
                if item_name:
                    string += scheme_name
                    first_index = string.find("%@")
                    second_index = string.find("%@", first_index + 1)
                    string = (
                        string[:first_index]
                        + str(item.number)
                        + " "
                        + item_name
                        + string[second_index + 2 :]
                    )
            string = string.replace("<br>", "\n\t")
            options.append(string)

        choice = dialog_creator.ChoiceInput.from_reduced(
            ["gain_scheme_items", "remove_scheme_items"],
            dialog="gain_remove_scheme_items",
        ).single_choice()
        if choice is None:
            return

        choice -= 1

        if choice == 0:
            self.add_scheme_items(options, scheme_items)
        elif choice == 1:
            self.remove_scheme_items(options, scheme_items)

    def add_scheme_items(
        self,
        options: list[str],
        scheme_items: dict[int, SchemeDataItem],
    ):
        scheme_ids, _ = dialog_creator.ChoiceInput.from_reduced(
            options,
            dialog="scheme_items_select_gain",
        ).multiple_choice()
        if scheme_ids is None:
            return
        for option_id in scheme_ids:
            scheme_id = list(scheme_items.keys())[option_id]
            if scheme_id not in self.to_obtain:
                self.to_obtain.append(scheme_id)
            if scheme_id in self.received:
                self.received.remove(scheme_id)

        color.ColoredText.localize("scheme_items_edit_success")

    def remove_scheme_items(
        self,
        options: list[str],
        scheme_items: dict[int, SchemeDataItem],
    ):
        scheme_ids, _ = dialog_creator.ChoiceInput.from_reduced(
            options,
            dialog="scheme_items_select_remove",
        ).multiple_choice()
        if scheme_ids is None:
            return
        for option_id in scheme_ids:
            scheme_id = list(scheme_items.keys())[option_id]
            if scheme_id in self.to_obtain:
                self.to_obtain.remove(scheme_id)
            if scheme_id in self.received:
                self.received.remove(scheme_id)

        color.ColoredText.localize("scheme_items_edit_success")



from __future__ import annotations
from typing import Any
from bcps import core
from bcps.cli import color


class Stage:
    def __init__(self, clear_times: int):
        self.clear_times = clear_times

    @staticmethod
    def init() -> Stage:
        return Stage(0)

    @staticmethod
    def read(data: core.Data) -> Stage:
        clear_times = data.read_short()
        return Stage(clear_times)

    def write(self, data: core.Data):
        data.write_short(self.clear_times)

    def serialize(self) -> int:
        return self.clear_times

    @staticmethod
    def deserialize(data: int) -> Stage:
        return Stage(
            data,
        )

    def __repr__(self):
        return f"Stage({self.clear_times})"

    def __str__(self):
        return self.__repr__()

    def clear_stage(self, clear_count: int = 1):
        self.clear_times = clear_count


class Chapter:
    def __init__(self, current_stage: int, total_stages: int = 0):
        self.current_stage = current_stage
        self.stages: list[Stage] = [Stage.init() for _ in range(total_stages)]

    @staticmethod
    def init(total_stages: int) -> Chapter:
        return Chapter(0, total_stages)

    @staticmethod
    def read_current_stage(data: core.Data):
        current_stage = data.read_byte()
        return Chapter(current_stage)

    def write_current_stage(self, data: core.Data):
        data.write_byte(self.current_stage)

    def read_stages(self, data: core.Data, total_stages: int):
        self.stages = [Stage.read(data) for _ in range(total_stages)]

    def write_stages(self, data: core.Data):
        for stage in self.stages:
            stage.write(data)

    def serialize(self) -> dict[str, Any]:
        return {
            "current_stage": self.current_stage,
            "stages": [stage.serialize() for stage in self.stages],
        }

    @staticmethod
    def deserialize(data: dict[str, Any]) -> Chapter:
        chapter = Chapter(data.get("current_stage", 0))
        chapter.stages = [Stage.deserialize(stage) for stage in data.get("stages", [])]
        return chapter

    def __repr__(self):
        return f"Chapter({self.current_stage}, {self.stages})"

    def __str__(self):
        return self.__repr__()


class ChaptersStars:
    def __init__(self, chapters: list[Chapter]):
        self.chapters = chapters

    @staticmethod
    def init(total_stages: int, total_stars: int) -> ChaptersStars:
        return ChaptersStars([Chapter.init(total_stages) for _ in range(total_stars)])

    @staticmethod
    def read_current_stage(data: core.Data, total_stars: int):
        chapters = [Chapter.read_current_stage(data) for _ in range(total_stars)]
        return ChaptersStars(chapters)

    def write_current_stage(self, data: core.Data):
        for chapter in self.chapters:
            chapter.write_current_stage(data)

    def read_stages(self, data: core.Data, total_stages: int):
        for chapter in self.chapters:
            chapter.read_stages(data, total_stages)

    def write_stages(self, data: core.Data):
        for chapter in self.chapters:
            chapter.write_stages(data)

    def serialize(self) -> list[dict[str, Any]]:
        return [chapter.serialize() for chapter in self.chapters]

    @staticmethod
    def deserialize(data: list[dict[str, Any]]) -> ChaptersStars:
        chapters = [Chapter.deserialize(chapter) for chapter in data]
        return ChaptersStars(chapters)

    def __repr__(self):
        return f"ChaptersStars({self.chapters})"

    def __str__(self):
        return self.__repr__()


class AkuChapters:
    def __init__(self, chapters: list[ChaptersStars]):
        self.chapters = chapters

    @staticmethod
    def init() -> AkuChapters:
        return AkuChapters([])

    @staticmethod
    def read(data: core.Data) -> AkuChapters:
        total_chapters = data.read_short()
        total_stages = data.read_byte()
        total_stars = data.read_byte()

        chapters = [
            ChaptersStars.read_current_stage(data, total_stars)
            for _ in range(total_chapters)
        ]

        for chapter in chapters:
            chapter.read_stages(data, total_stages)

        return AkuChapters(chapters)

    def write(self, data: core.Data):
        data.write_short(len(self.chapters))
        try:
            data.write_byte(len(self.chapters[0].chapters[0].stages))
        except IndexError:
            data.write_byte(0)
        try:
            data.write_byte(len(self.chapters[0].chapters))
        except IndexError:
            data.write_byte(0)

        for chapter in self.chapters:
            chapter.write_current_stage(data)

        for chapter in self.chapters:
            chapter.write_stages(data)

    def serialize(self) -> list[list[dict[str, Any]]]:
        return [chapter.serialize() for chapter in self.chapters]

    @staticmethod
    def deserialize(data: list[list[dict[str, Any]]]) -> AkuChapters:
        chapters = [ChaptersStars.deserialize(chapter) for chapter in data]
        return AkuChapters(chapters)

    def __repr__(self):
        return f"Chapters({self.chapters})"

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def edit_aku_chapters(save_file: core.SaveFile):
        aku = save_file.aku
        chapter = aku.chapters[0].chapters[0]

        clear_progress = core.StoryChapters.get_selected_chapter_progress(
            max_stages=len(chapter.stages)
        )
        if clear_progress is None:
            return

        if clear_progress > 1:
            individual_clear_count = core.StoryChapters.ask_if_individual_clear_counts()
            if individual_clear_count is None:
                return
        else:
            individual_clear_count = True

        if individual_clear_count:
            stage_names = core.StageNames(save_file, "DM", 49).stage_names
            if stage_names is None:
                return
            for i, stage in enumerate(chapter.stages[:clear_progress]):
                stage_name = stage_names[i]
                color.ColoredText.localize("aku_current_stage", name=stage_name, id=i)
                clear_count = core.StoryChapters.ask_clear_count()
                if clear_count is None:
                    return
                stage.clear_stage(clear_count)
        else:
            clear_count = core.StoryChapters.ask_clear_count()
            if clear_count is None:
                return
            for stage in chapter.stages[:clear_progress]:
                stage.clear_stage(clear_count)

        for i in range(clear_progress, len(chapter.stages)):
            chapter.stages[i].clear_stage(clear_count=0)

        color.ColoredText.localize("aku_clear_success")
